# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for Model objects used in Python SDK."""

import os
from collections import deque
from pathlib import Path
from typing import Deque, List, Tuple


class DirectoryTree(dict):
    r"""Create a safe directory tree from paths.

    Example usage:

    >>> directory = DirectoryTree()
    >>> directory.add('a/b/c')
    >>> directory.add('a/b/c/d')
    >>> directory.add('x/y/z')
    >>> directory.add('x/y/zz')
    >>> print('\n'.join(sorted(directory)))
    a/b/c/d
    x/y/z
    x/y/zz
    >>> print('\n'.join(sorted(directory.get('x/y'))))
    z
    zz

    """

    @classmethod
    def from_list(cls, values):
        """Construct a tree from a list with paths."""
        self = cls()
        for value in values:
            self.add(value)
        return self

    def get(self, value, default=None):
        """Return a subtree if exists."""
        path = value if isinstance(value, Path) else Path(str(value))
        subtree = self
        for part in path.parts:
            try:
                subtree = subtree[part]
            except KeyError:
                return default
        return subtree

    def add(self, value):
        """Create a safe directory from a value."""
        path = value if isinstance(value, Path) else Path(str(value))
        if path and path != path.parent:
            destination = self
            for part in path.parts:
                destination = destination.setdefault(part, DirectoryTree())

    def __iter__(self):
        """Yield all stored directories."""
        filter = {
            os.path.sep,
        }
        queue: Deque[Tuple["DirectoryTree", List[str]]] = deque()
        queue.append((self, []))

        while queue:
            data, parents = queue.popleft()
            for key, value in dict.items(data):
                if key in filter:
                    continue
                if value:
                    queue.append((value, parents + [key]))
                else:
                    yield os.path.sep.join(parents + [key])



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset add business logic."""

import itertools
import os.path
import shutil
import tempfile
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Union, cast

from renku.command.command_builder.command import inject
from renku.core import errors
from renku.core.dataset.context import DatasetContext
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.dataset.pointer_file import create_external_file
from renku.core.dataset.providers.api import ImporterApi
from renku.core.dataset.providers.factory import ProviderFactory
from renku.core.dataset.providers.local import LocalProvider
from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.interface.storage import IStorage
from renku.core.storage import check_external_storage, track_paths_in_storage
from renku.core.util import communication, requests
from renku.core.util.git import get_git_user
from renku.core.util.os import get_absolute_path, get_file_size, get_files, get_relative_path, hash_file, is_subpath
from renku.core.util.urls import check_url, is_uri_subfolder, resolve_uri
from renku.core.util.util import parallel_execute
from renku.domain_model.constant import NON_EXISTING_ENTITY_CHECKSUM
from renku.domain_model.dataset import Dataset, DatasetFile, RemoteEntity
from renku.domain_model.project_context import project_context


def add_to_dataset(
    dataset_name: str,
    urls: List[str],
    *,
    importer: Optional[ImporterApi] = None,
    force: bool = False,
    create: bool = False,
    overwrite: bool = False,
    sources: Optional[List[Union[str, Path]]] = None,
    destination: str = "",
    revision: Optional[str] = None,
    extract: bool = False,
    clear_files_before: bool = False,
    total_size: Optional[int] = None,
    datadir: Optional[Path] = None,
    storage: Optional[str] = None,
    **kwargs,
) -> Dataset:
    """Import the data into the data directory."""
    sources = sources or []

    if not create and storage:
        raise errors.ParameterError("Storage can be set only when creating a dataset")

    try:
        with DatasetContext(name=dataset_name, create=create, datadir=datadir, storage=storage) as dataset:
            destination_path = _create_destination_directory(dataset, destination)

            check_external_storage()

            # NOTE: This is not required for cloud storages
            if not dataset.storage:
                _check_available_space(urls, total_size=total_size)

            datadir = cast(Path, project_context.path / dataset.get_datadir())
            if create and datadir.exists() and not dataset.storage:
                # NOTE: Add datadir to paths to add missing files on create
                for file in get_files(datadir):
                    urls.append(str(file))

            files = get_files_metadata(
                urls=urls,
                dataset=dataset,
                importer=importer,
                destination=destination_path,
                revision=revision,
                sources=sources,
                extract=extract,
                force=force,
                **kwargs,
            )

            files = filter_files(dataset=dataset, files=files, force=force, overwrite=overwrite)
            if not files:
                if create:
                    raise errors.UsageError("There are no files to create a dataset")
                else:
                    communication.warn("No new file was added to project")
                    return dataset

            # NOTE: All files at this point can be force-added

            copy_files_to_dataset(dataset=dataset, files=files)
            add_files_to_repository(dataset=dataset, files=files)
            update_dataset_metadata(dataset=dataset, files=files, clear_files_before=clear_files_before)

            DatasetsProvenance().add_or_update(dataset, creator=get_git_user(repository=project_context.repository))
    except errors.DatasetNotFound:
        raise errors.DatasetNotFound(
            message="Dataset '{0}' does not exist.\n"
            "Use 'renku dataset create {0}' to create the dataset or retry 'renku dataset add {0}' command "
            "with '--create' option for automatic dataset creation.".format(dataset_name)
        )
    except (FileNotFoundError, errors.GitCommandError) as e:
        raise errors.ParameterError("Could not find paths/URLs: \n{}".format("\n".join(urls))) from e
    else:
        project_context.database.commit()
        return dataset


def get_files_metadata(
    *,
    urls: List[str],
    importer: Optional[ImporterApi] = None,
    dataset: Dataset,
    destination: Path,
    extract: bool,
    revision: Optional[str],
    sources: List[Union[str, Path]],
    force: bool = False,
    **kwargs,
) -> List[DatasetAddMetadata]:
    """Process file URLs for adding to a dataset."""
    if importer:
        return importer.download_files(destination=destination, extract=extract)

    if len(urls) == 0:
        raise errors.ParameterError("No URL is specified")
    if sources and len(urls) > 1:
        raise errors.ParameterError("Cannot use '--source' with multiple URLs.")

    files = []

    for url in urls:
        is_remote, is_git = check_url(url)

        if not is_git and sources:
            raise errors.ParameterError("Cannot use '-s/--src/--source' with URLs or local files.")

        dataset_add_action = DatasetAddAction.NONE

        if is_remote:
            provider = ProviderFactory.get_add_provider(uri=url)
        else:
            # NOTE: If URI is in the local file system, check to see if it's part of a mounted dataset/provider
            cloud_dataset, remote_url = get_cloud_dataset_from_path(path=url)
            if cloud_dataset:
                url = remote_url
                provider = ProviderFactory.get_storage_provider(uri=cloud_dataset.storage)
                # NOTE: Update metadata if destination dataset is the same as source dataset, otherwise copy the file
                # since it's already in the local filesystem
                dataset_add_action = DatasetAddAction.COPY
            else:
                provider = LocalProvider(uri=url)

        new_files = provider.get_metadata(
            uri=url,
            destination=destination,
            revision=revision,
            sources=sources,
            dataset=dataset,
            extract=extract,
            force=force,
            dataset_add_action=dataset_add_action,
            **kwargs,
        )

        files.extend(new_files)

    return files


@inject.autoparams("dataset_gateway")
def has_cloud_storage(dataset_gateway: IDatasetGateway) -> bool:
    """Return if a project has any dataset with cloud storage with its data directory mounted or pulled."""
    # NOTE: ``exists`` return False for symlinks if their target doesn't exists, but it's fine here since it means the
    # dataset's mounted/pulled location doesn't exist.
    return any(
        dataset
        for dataset in dataset_gateway.get_all_active_datasets()
        if dataset.storage and (project_context.path / dataset.get_datadir()).exists()
    )


@inject.autoparams("dataset_gateway")
def get_cloud_dataset_from_path(
    path: Union[Path, str], dataset_gateway: IDatasetGateway, missing_ok: bool = False
) -> Tuple[Optional[Dataset], Optional[str]]:
    """Check the path against datasets' storage and return a dataset (if any)."""
    if not has_cloud_storage():
        return None, None

    # NOTE: If path is inside the datadir of a dataset with storage backend and the dataset isn't mounted, we should
    # still add whatever is in the path (because it might have been pulled)

    path = Path(get_absolute_path(path))

    if not missing_ok and not path.exists() and not os.path.lexists(path):
        return None, None

    for dataset in dataset_gateway.get_all_active_datasets():
        if not dataset.storage:
            continue

        datadir = project_context.path / dataset.get_datadir()
        resolved_path = path.resolve()

        # NOTE: Resolve ``path`` because ``datadir`` is resolved and resolved paths might have be on a different
        # location (e.g. on macos /tmp resolves to /private/tmp)
        resolved_relative_path = get_relative_path(resolved_path, base=datadir.resolve())

        if is_subpath(path, base=datadir) or resolved_relative_path is not None:
            if resolved_relative_path == ".":
                resolved_relative_path = ""
            storage_uri = dataset.storage.rstrip("/")
            remote_url = f"{storage_uri}/{resolved_relative_path}"
            return dataset, remote_url
        elif is_subpath(resolved_path, Path(dataset.storage).resolve()):  # NOTE: For local backend storage
            return dataset, str(resolved_path)

    return None, None


def _check_available_space(urls: List[str], total_size: Optional[int] = None):
    """Check that there is enough space available on the device for download."""
    if total_size is None:
        total_size = 0
        for url in urls:
            is_remote, _ = check_url(url)
            if not is_remote:
                continue

            try:
                response = requests.head(url, allow_redirects=True)
                total_size += int(response.headers.get("content-length", 0))
            except errors.RequestError:
                pass
    usage = shutil.disk_usage(project_context.path)

    if total_size > usage.free:
        mb = 2**20
        message = "Insufficient disk space (required: {:.2f} MB" "/available: {:.2f} MB). ".format(
            total_size / mb, usage.free / mb
        )
        raise errors.OperationError(message)


def _create_destination_directory(dataset: Dataset, destination: Optional[Union[Path, str]] = None) -> Path:
    """Create directory for dataset add."""
    dataset_datadir = project_context.path / dataset.get_datadir()

    # NOTE: Make sure that dataset's data dir exists because we check for existence of a destination later to decide
    # what will be its name
    dataset_datadir.mkdir(parents=True, exist_ok=True)

    destination = destination or ""
    relative_path = cast(str, get_relative_path(destination, base=dataset_datadir, strict=True))
    return dataset_datadir / relative_path


def filter_files(
    dataset: Dataset, files: List[DatasetAddMetadata], force: bool, overwrite: bool
) -> List[DatasetAddMetadata]:
    """Filter ignored and overwritten files."""

    def remove_git_files(files_to_filter: List[DatasetAddMetadata]):
        """Remove all files that are under a .git directory."""
        git_paths = [f.entity_path for f in files_to_filter if str(f.entity_path).startswith(".git")]
        if not git_paths:
            return files_to_filter

        communication.warn("Ignored adding paths under a .git directory:\n\t" + "\n\t".join(str(p) for p in git_paths))
        return [f for f in files_to_filter if f.entity_path not in git_paths]

    def check_ignored_files(files_to_filter: Generator[DatasetAddMetadata, None, None]):
        """Check if any files added were ignored."""
        paths = {f.get_absolute_commit_path(project_context.path): f for f in files_to_filter}

        ignored_paths = project_context.repository.get_ignored_paths(*paths)
        if ignored_paths:
            ignored_sources = [file.source for path, file in paths.items() if path in ignored_paths]

            communication.warn(
                "Theses paths are ignored by one of your .gitignore files (use '--force' flag if you really want to "
                "add them):\n\t" + "\n\t".join([str(p) for p in ignored_sources])
            )

        return (file for path, file in paths.items() if path not in ignored_paths)

    def check_existing_files(files_to_filter: Generator[DatasetAddMetadata, None, None]):
        """Check if files added already exist."""
        files_list = list(files_to_filter)
        existing_paths = [f.entity_path for f in files_list if dataset.find_file(f.entity_path)]
        if existing_paths:
            communication.warn(
                "These existing files were not overwritten (use '--overwrite' flag to overwrite them):\n\t"
                + "\n\t".join([str(p) for p in existing_paths])
            )

        return (f for f in files_list if f.entity_path not in existing_paths)

    files = remove_git_files(files)

    # NOTE: Don't filter ignored or existing files that will be added to a remote storage
    remote_files = (f for f in files if f.metadata_only)
    local_files = (f for f in files if not f.metadata_only)

    # NOTE: Data directory of datasets with a storage backend is always ignored, so, filtering files is meaningless
    if not force and not dataset.storage:
        local_files = check_ignored_files(local_files)

    if not overwrite:
        local_files = check_existing_files(local_files)

    files = list(itertools.chain(local_files, remote_files))

    return files


def get_dataset_file_path_within_dataset(dataset: Dataset, entity_path: Union[Path, str]) -> Path:
    """Return a dataset file's path relative to the dataset's datadir.

    NOTE: Dataset files can reside outside the dataset's datadir.
    """
    assert not os.path.isabs(entity_path), f"Entity path cannot be absolute: {entity_path}"

    entity_path = Path(entity_path)

    try:
        return entity_path.relative_to(dataset.get_datadir())
    except ValueError:
        return entity_path


def get_upload_uri(dataset: Dataset, entity_path: Union[Path, str]) -> str:
    """Return the remote storage path that a dataset file would be located.

    Args:
        dataset(Dataset): Dataset with a backend storage.
        entity_path(Union[Path, str]): Dataset file's path (entity path); it is relative to the project's root.

    Returns:
        str: URI within remote storage.
    """
    assert dataset.storage, "Cannot get URI for datasets with no backend storage"

    base = dataset.storage.rstrip("/")
    path_within_dataset = get_dataset_file_path_within_dataset(dataset=dataset, entity_path=entity_path)

    return f"{base}/{path_within_dataset}"


def copy_file(file: DatasetAddMetadata, dataset: Dataset, storage: Optional[IStorage]) -> List[Optional[Path]]:
    """Copy/move/link a file to dataset's data directory."""
    if not file.has_action:
        return []

    # NOTE: If file is in a subdirectory of a dataset's remote storage URI, only update the metadata
    if file.from_cloud_storage:
        if dataset.storage and is_uri_subfolder(resolve_uri(dataset.storage), file.url):
            file.action = DatasetAddAction.METADATA_ONLY
        else:
            file.action = DatasetAddAction.DOWNLOAD

    if file.action in (
        DatasetAddAction.COPY,
        DatasetAddAction.MOVE,
        DatasetAddAction.SYMLINK,
        DatasetAddAction.DOWNLOAD,
    ):
        try:
            file.destination.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            raise errors.InvalidFileOperation(f"Cannot create destination '{file.destination.parent}': {e}")

    file_to_upload: Union[Path, str] = file.source.resolve()
    delete_source = False
    track_in_lfs = True

    try:
        if file.action == DatasetAddAction.DOWNLOAD:
            # NOTE: Download to a temporary location if dataset has a cloud storage because it's usually mounted as
            # read-only and download would fail. It's ok not to move it to dataset's data dir since it'll be uploaded.
            dst: Union[Path, str]
            if storage:
                fd, dst = tempfile.mkstemp()
                os.close(fd)
            else:
                dst = file.destination

            assert file.provider, f"Storage provider isn't set for {file} with DOWNLOAD action"
            download_storage = file.provider.get_storage()
            download_storage.download(file.url, dst)
            file_to_upload = dst
        elif file.action == DatasetAddAction.COPY:
            shutil.copy(file.source, file.destination)
        elif file.action == DatasetAddAction.MOVE:
            # NOTE: Set ``delete_source`` in case move fails due to a dataset's read-only mounted data directory
            delete_source = True
            shutil.move(file.source, file.destination, copy_function=shutil.copy)  # type: ignore
            delete_source = False
            file_to_upload = file.destination
        elif file.action == DatasetAddAction.SYMLINK:
            create_external_file(target=file.source, path=file.destination)
            # NOTE: Don't track symlinks to external files in LFS
            track_in_lfs = False
        elif file.metadata_only:
            # NOTE: Nothing to do when adding file to a dataset with a parent remote storage
            pass
        else:
            raise errors.OperationError(f"Invalid action {file.action}")
    except OSError as e:
        # NOTE: It's ok if copying data to a read-only mounted cloud storage fails
        if "Read-only file system" in str(e) and storage:
            pass
        else:
            dst = get_relative_path(file.destination, project_context.path) or file.destination
            raise errors.InvalidFileOperation(f"Cannot copy/move '{dst}': {e}")

    if file.size is None:
        file.size = get_file_size(file_to_upload)

    if storage:
        # NOTE: Don't track files in a dataset with cloud storage in LFS
        track_in_lfs = False

        if file.metadata_only:
            assert file.based_on, f"wasBasedOn isn't set for {file} with METADATA_ONLY action"
            file_uri = file.based_on.url
            md5_hash: Optional[str] = file.based_on.checksum
        else:
            file_uri = get_upload_uri(dataset=dataset, entity_path=file.entity_path)
            md5_hash = hash_file(file_to_upload, hash_type="md5")

            # NOTE: If dataset has a storage backend, upload the file to the remote storage.
            storage.upload(source=file_to_upload, uri=file_uri)

        file.based_on = RemoteEntity(url=file_uri, path=file.entity_path, checksum=md5_hash)

    if delete_source:
        file.source.unlink(missing_ok=True)

    return [file.destination] if track_in_lfs else []


def copy_files_to_dataset(dataset: Dataset, files: List[DatasetAddMetadata]):
    """Copy/Move files into a dataset's directory."""

    dataset_storage = None
    if dataset.storage:
        provider = ProviderFactory.get_storage_provider(uri=dataset.storage)
        dataset_storage = provider.get_storage()

    lfs_files = parallel_execute(copy_file, files, rate=5, dataset=dataset, storage=dataset_storage)

    if lfs_files and not dataset.storage:
        track_paths_in_storage(*lfs_files)


def add_files_to_repository(dataset: Dataset, files: List[DatasetAddMetadata]):
    """Track files in project's repository."""
    # NOTE: There is nothing to track for remote storages
    if dataset.storage:
        return

    # NOTE: Don't commit files that will be uploaded to a remote storage
    paths_to_commit = [f.get_absolute_commit_path(project_context.path) for f in files]

    repository = project_context.repository

    # Force-add to include possible ignored files
    if len(paths_to_commit) > 0:
        repository.add(*paths_to_commit, project_context.pointers_path, force=True)

    n_staged_changes = len(repository.staged_changes)
    if n_staged_changes == 0:
        communication.warn("No new file was added to project")


def update_dataset_metadata(dataset: Dataset, files: List[DatasetAddMetadata], clear_files_before: bool):
    """Add newly-added files to the dataset's metadata."""
    # NOTE: For datasets with cloud storage backend, we use MD5 hash as checksum instead of git hash.
    if dataset.storage:
        checksums: Dict[Union[Path, str], Optional[str]] = {
            f.entity_path: f.based_on.checksum for f in files if f.based_on
        }
    else:
        repo_paths: List[Union[Path, str]] = [
            file.entity_path for file in files if (project_context.path / file.entity_path).exists()
        ]
        checksums = project_context.repository.get_object_hashes(repo_paths)

    dataset_files = []

    for file in files:
        dataset_file = DatasetFile.from_path(
            path=file.entity_path,
            source=file.url,
            based_on=file.based_on,
            size=file.size,
            checksum=checksums.get(file.entity_path) or NON_EXISTING_ENTITY_CHECKSUM,
        )
        dataset_files.append(dataset_file)

    if clear_files_before:
        dataset.clear_files()

    dataset.add_or_update_files(dataset_files)



#
# Copyright 2018-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Rollback project to a previous point in time.

Description
~~~~~~~~~~~

Undo actions taken using Renku in a project.
This command shows a list of all actions done by renku and lets you pick
one that you want to return to, discarding any changes done in the repo (by
Renku or manually) after that point.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.rollback:rollback
   :prog: renku rollback
   :nested: full

Examples
~~~~~~~~

Once you pick a checkpoint to return to,
the commands shows all files and Renku objects that would be affected by the
rollback and how they would be affected. If you confirm, the project will be
reset to that point in time, with anything done after that point being
deleted/lost.

.. code-block:: console

    $ renku rollback
    Select a checkpoint to roll back to:

    [0] 2021-10-20 09:50:04         renku workflow edit cp-blabla-asdasf-0b535 --name test
    [1] 2021-10-20 09:49:19         renku rerun asdasf
    [2] 2021-10-20 09:48:59         renku run cp blabla asdasf
    [3] 2021-10-20 08:37:00         renku dataset add e blabla
    [4] 2021-10-20 08:31:16         renku dataset create m
    Checkpoint ([q] to quit) [q]: 4
    The following changes would be done:

    Metadata:

        Modified ♻️:
            Dataset: e

        Removed 🔥:
            Plan: cp-blabla-asdasf-0b535
            Plan: test
            Run: /activities/cc3ab70952fc499e93e7e4075a076bf5 (Plan name: cp-blabla-asdasf-0b535)
            Run: /activities/48b89b22567d4282abe8a016fa91878f (Plan name: cp-blabla-asdasf-0b535)

    Files:

        Restored ↻:
            blabla

        Removed 🔥:
            asdasf

    Proceed? [y/N]: y

.. note:: This command was introduced in renku-python version 1.0.0. Commands
   executed with previous versions of renku can't be rolled back to.

.. cheatsheet::
   :group: Undo Renku Command
   :command: $ renku rollback
   :description: Rollback project to a previous point in time.
   :target: rp,ui

"""

import click

from renku.ui.cli.utils.callback import ClickCallback


@click.command()
@click.pass_context
def rollback(ctx):
    """Rollback project to a previous point in time.

    Only renku commands executed after the 1.0.0 release are supported,
    previous commands won't show up.
    """
    from renku.command.rollback import rollback_command

    communicator = ClickCallback()
    rollback_command().with_communicator(communicator).build().execute()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom git merge tool for renku metadata.

Description
~~~~~~~~~~~

Renku stores all of its metadata in the repository, in compressed form.
When working in multiple branches with Renku, this metadata needs to be merged
when a git merge is made. To support users when doing this, Renku provides a
custom merge tool that takes care of merging the metadata.

The merge tool is set up automatically when creating a new project or when
using ``renku clone`` to clone a Renku project.
Keep in mind the merge tool configuration cannot be shared through remote
repositories and has to be set up on each clone separately.

You can manually set up the merge tool by running ``renku mergetool install``.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.mergetool:mergetool
   :prog: renku mergetool
   :nested: full
"""

import click

from renku.ui.cli.utils.callback import ClickCallback


@click.group()
def mergetool():
    """Mergetool commands."""
    pass


@mergetool.command(hidden=True)
@click.argument("base", type=click.Path(exists=True))
@click.argument("local", type=click.Path(exists=True))
@click.argument("remote", type=click.Path(exists=True))
def merge(base, local, remote):
    """Move files and check repository for potential problems."""
    from renku.command.mergetool import mergetool_command

    communicator = ClickCallback()
    mergetool_command().with_communicator(communicator).build().execute(local=local, remote=remote, base=base)


@mergetool.command()
def install():
    """Setup mergetool locally."""
    from renku.command.mergetool import mergetool_install_command

    mergetool_install_command().with_commit(commit_only=[".gitattributes"]).require_clean().build().execute()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for dataset list files."""

import re
from subprocess import PIPE, SubprocessError, run
from typing import Callable, Dict, Optional, Union

from renku.domain_model.project_context import project_context


def tabular(records, *, columns=None):
    """Format dataset files with a tabular output.

    Args:
        records: Filtered collection.
        columns: List of columns to display (Default value = None)

    Returns:
        String of records in tabular representation.
    """
    from renku.command.format.tabulate import tabulate

    if not columns:
        columns = "added,creators,dataset,full_path"

    for record in records:
        record.creators = record.dataset.creators

    return tabulate(
        collection=records,
        columns=columns,
        columns_mapping=DATASET_FILES_COLUMNS,
        columns_alignments=DATASET_FILES_COLUMNS_ALIGNMENTS,
    )


def get_lfs_tracking_and_file_sizes(records, has_tag: bool):
    """Try to get file size from Git LFS and check if files are tracked in git lfs.

    Args:
        records: File records tog et size for.
        has_tag(bool): Whether sizes are retrieved for a given tag instead of HEAD commit
    """
    from humanize import naturalsize  # Slow import

    repository = project_context.repository

    def get_lfs_tracking():
        if has_tag:
            return

        paths = (r.path for r in records)
        attrs = repository.get_attributes(*paths)

        for record in records:
            if attrs.get(str(record.path), {}).get("filter") == "lfs":
                record.is_lfs = True
            else:
                record.is_lfs = False

    def naturalize(value) -> str:
        return naturalsize(value).upper().replace("BYTES", " B")

    def get_file_sizes():
        if not any(r for r in records if r.size is None):  # All records already have a size
            return {}, {}

        lfs_files_sizes = {}

        try:
            lfs_run = run(
                ("git", "lfs", "ls-files", "--name-only", "--size", "--deleted"),
                stdout=PIPE,
                cwd=project_context.path,
                text=True,
            )
        except SubprocessError:
            pass
        else:
            lfs_output = lfs_run.stdout.split("\n")
            # Example line format: relative/path/to/file (7.9 MB)
            pattern = re.compile(r"^(.*?)\s*\((.*)\)")

            for line in lfs_output:
                match = pattern.search(line)
                if not match:
                    continue
                path, size = match.groups()
                # Fix alignment for bytes
                if size.endswith(" B"):
                    size = size.replace(" B", "  B")
                lfs_files_sizes[path] = size

        non_lfs_files_sizes: Dict[str, Optional[Union[int, str]]]
        if has_tag:
            checksums = [r.entity.checksum for r in records]
            sizes = repository.get_sizes(*checksums)
            non_lfs_files_sizes = {k.entity.path: naturalize(v) for k, v in zip(records, sizes)}
        else:
            non_lfs_files_sizes = {
                o.path: o.size for o in repository.head.commit.traverse() if o.path not in lfs_files_sizes
            }
            non_lfs_files_sizes = {k: naturalize(v) for k, v in non_lfs_files_sizes.items()}

        return lfs_files_sizes, non_lfs_files_sizes

    lfs_files_sizes, non_lfs_files_sizes = get_file_sizes()
    get_lfs_tracking()

    for record in records:
        size = (
            lfs_files_sizes.get(record.path) or non_lfs_files_sizes.get(record.path) or None
            if record.size is None
            else naturalize(record.size)
        )
        record.size = size if size or size == 0 else None

        # NOTE: When listing a tag we assume that the file is in LFS if it was in LFS at some point in time
        if has_tag:
            record.is_lfs = lfs_files_sizes.get(record.path) is not None


def jsonld(records, **_):
    """Format dataset files as JSON-LD.

    Args:
        records: Filtered collection.
    """
    from renku.command.format.json import dumps
    from renku.command.schema.dataset import DatasetFileSchema

    data = [DatasetFileSchema(flattened=True).dump(record) for record in records]
    return dumps(data, indent=2)


def json(records, **_):
    """Format dataset files as JSON.

    Args:
        records: Filtered collection.

    Returns:
        String of records in JSON representation.
    """
    from renku.command.format.json import dumps
    from renku.domain_model.dataset import DatasetFileDetailsJson

    for record in records:
        record.creators = record.dataset.creators

    data = [DatasetFileDetailsJson().dump(record) for record in records]
    return dumps(data, indent=2)


DATASET_FILES_FORMATS: Dict[str, Callable] = {
    "tabular": tabular,
    "json-ld": jsonld,
    "json": json,
}
"""Valid formatting options."""

DATASET_FILES_COLUMNS = {
    "added": ("date_added", "added"),
    "checksum": ("entity.checksum", "checksum"),
    "creators": ("creators_csv", "creators"),
    "creators_full": ("creators_full_csv", "creators"),
    "dataset": ("title", "dataset"),
    "full_path": ("full_path", None),
    "path": ("path", None),
    "short_name": ("dataset_name", "dataset name"),
    "dataset_name": ("dataset_name", "dataset name"),
    "size": ("size", None),
    "lfs": ("is_lfs", "lfs"),
    "source": ("source", None),
}

DATASET_FILES_COLUMNS_ALIGNMENTS = {"size": "right"}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent a group of run templates."""

from marshmallow import EXCLUDE, pre_dump

from renku.command.schema.agent import PersonSchema
from renku.command.schema.calamus import JsonLDSchema, Nested, fields, prov, renku, schema
from renku.command.schema.parameter import ParameterLinkSchema, ParameterMappingSchema
from renku.command.schema.plan import PlanSchema
from renku.domain_model.workflow.composite_plan import CompositePlan


class CompositePlanSchema(JsonLDSchema):
    """Plan schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Plan, schema.Action, schema.CreativeWork, renku.CompositePlan]
        model = CompositePlan
        unknown = EXCLUDE

    description = fields.String(schema.description, load_default=None)
    id = fields.Id()
    creators = Nested(schema.creator, PersonSchema, many=True)
    mappings = Nested(renku.hasMappings, [ParameterMappingSchema], many=True, load_default=None)
    date_created = fields.DateTime(schema.dateCreated, format="iso")
    date_modified = fields.DateTime(schema.dateModified, format="iso")
    date_removed = fields.DateTime(prov.invalidatedAtTime, format="iso")
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)
    name = fields.String(schema.name, load_default=None)
    derived_from = fields.IRI(prov.wasDerivedFrom, load_default=None)
    project_id = fields.IRI(renku.hasPlan, reverse=True)
    plans = Nested(renku.hasSubprocess, [PlanSchema, "CompositePlanSchema"], many=True)
    links = Nested(renku.workflowLinks, [ParameterLinkSchema], many=True, load_default=None)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            obj.date_created = obj.date_created.replace(microsecond=0)
            obj.date_modified = obj.date_modified.replace(microsecond=0)
            if obj.date_removed:
                obj.date_removed = obj.date_removed.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku management dataset request models."""

import imghdr
import os
import shutil
import urllib
from pathlib import Path
from typing import List, Optional, Union, cast
from urllib.request import urlretrieve

from renku.core import errors
from renku.domain_model.dataset import Dataset, ImageObject
from renku.domain_model.project_context import project_context


class ImageRequestModel:
    """Model for passing image information to dataset use-cases."""

    def __init__(
        self,
        content_url: str,
        position: int,
        mirror_locally: bool = False,
        safe_image_paths: Optional[List[str]] = None,
    ) -> None:
        self.content_url = content_url
        self.position = position
        self.mirror_locally = mirror_locally
        self.safe_image_paths: List[Union[str, Path]] = cast(List[Union[str, Path]], safe_image_paths) or []

    def to_image_object(self, dataset: Dataset) -> ImageObject:
        """Convert request model to ``ImageObject``."""
        image_type = None
        self.safe_image_paths.append(project_context.path)

        image_folder = project_context.dataset_images_path / dataset.initial_identifier
        image_folder.mkdir(exist_ok=True, parents=True)

        if urllib.parse.urlparse(self.content_url).netloc:
            # NOTE: absolute url
            if not self.mirror_locally:
                return ImageObject(
                    content_url=self.content_url,
                    position=self.position,
                    id=ImageObject.generate_id(dataset_id=dataset.id, position=self.position),
                )

            # NOTE: mirror the image locally
            try:
                path, _ = urlretrieve(self.content_url)
            except urllib.error.URLError as e:
                raise errors.DatasetImageError(f"Dataset image with url {self.content_url} couldn't be mirrored") from e

            image_type = imghdr.what(path)
            if image_type:
                image_type = f".{image_type}"

            self.content_url = path
            self.safe_image_paths.append(Path(path).parent)

        path = self.content_url
        if not os.path.isabs(path):
            path = os.path.normpath(os.path.join(project_context.path, path))

        if not os.path.exists(path) or not any(
            os.path.commonprefix([path, p]) == str(p) for p in self.safe_image_paths
        ):
            # NOTE: make sure files exists and prevent path traversal
            raise errors.DatasetImageError(f"Dataset image with relative path {self.content_url} not found")

        if not path.startswith(str(image_folder)):
            # NOTE: only copy dataset image if it's not in .renku/datasets/<id>/images/ already
            if image_type:
                ext = image_type
            else:
                _, ext = os.path.splitext(self.content_url)

            img_path = image_folder / f"{self.position}{ext}"
            shutil.copy(path, img_path)
        else:
            img_path = Path(path)

        return ImageObject(
            content_url=str(img_path.relative_to(project_context.path)),
            position=self.position,
            id=ImageObject.generate_id(dataset_id=dataset.id, position=self.position),
        )



#  Copyright Swiss Data Science Center (SDSC). A partnership between
#  École Polytechnique Fédérale de Lausanne (EPFL) and
#  Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Docker based interactive session provider."""

import os
import platform
import webbrowser
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Union, cast
from uuid import uuid4

import docker
from requests.exceptions import ReadTimeout

from renku.core import errors
from renku.core.config import get_value
from renku.core.constant import ProviderPriority
from renku.core.plugin import hookimpl
from renku.core.util import communication
from renku.domain_model.project_context import project_context
from renku.domain_model.session import ISessionProvider, Session, SessionStopStatus

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderParameter


class DockerSessionProvider(ISessionProvider):
    """A docker based interactive session provider."""

    JUPYTER_PORT = 8888
    # NOTE: Give the docker provider the highest priority so that it's checked first
    priority: ProviderPriority = ProviderPriority.HIGHEST

    def __init__(self):
        self._docker_client = None

    def docker_client(self) -> docker.client.DockerClient:
        """Get the docker client.

        Note:
            This is not a @property, even though it should be, because ``pluggy``
            will call it in that case in unrelated parts of the code.
        Raises:
            errors.DockerError: Exception when docker is not available.
        Returns:
            The docker client.
        """
        if self._docker_client is None:
            try:
                self._docker_client = docker.from_env()
            except docker.errors.DockerException as e:
                raise errors.DockerError(f"Error while initializing docker client: {str(e)}")
        return self._docker_client

    @staticmethod
    def _get_jupyter_urls(ports: Dict[str, Any], auth_token: str, jupyter_port: int = 8888) -> Iterable[str]:
        port_key = f"{jupyter_port}/tcp"
        if port_key not in ports:
            return list()
        return map(lambda x: f"http://{x['HostIp']}:{x['HostPort']}/?token={auth_token}", ports[port_key])

    def _get_docker_containers(self, project_name: str) -> List[docker.models.containers.Container]:
        return self.docker_client().containers.list(filters={"label": f"renku_project={project_name}"})

    @property
    def name(self) -> str:
        """Return session provider's name."""
        return "docker"

    def is_remote_provider(self) -> bool:
        """Return True for remote providers (i.e. not local Docker)."""
        return False

    def build_image(self, image_descriptor: Path, image_name: str, config: Optional[Dict[str, Any]]):
        """Builds the container image."""
        self.docker_client().images.build(path=str(image_descriptor), tag=image_name)

    def find_image(self, image_name: str, config: Optional[Dict[str, Any]]) -> bool:
        """Find the given container image."""
        with communication.busy(msg=f"Checking for image {image_name}"):
            try:
                self.docker_client().images.get(image_name)
            except docker.errors.ImageNotFound:
                try:
                    self.docker_client().images.get_registry_data(image_name)
                except docker.errors.APIError:
                    return False
            else:
                return True

        try:
            with communication.busy(msg=f"Pulling image from remote {image_name}"):
                self.docker_client().images.pull(image_name)
        except docker.errors.NotFound:
            return False
        else:
            return True

    @hookimpl
    def session_provider(self) -> ISessionProvider:
        """Supported session provider.

        Returns:
            A reference to ``self``.
        """
        return self

    def get_start_parameters(self) -> List["ProviderParameter"]:
        """Returns parameters that can be set for session start."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter("port", help="Local port to use (random if not specified).", type=int),
            ProviderParameter("force-build", help="Always build image and don't check if it exists.", is_flag=True),
            ProviderParameter(
                "blkio-weight", help="Block IO (relative weight), between 10 and 1000, or 0 to disable.", type=int
            ),
            ProviderParameter("cap-add", help="Add Linux capabilities.", multiple=True),
            ProviderParameter("cap-drop", help="Drop Linux capabilities.", multiple=True),
            ProviderParameter("cgroup-parent", help="Override the default parent cgroup.", type=str),
            ProviderParameter("cpu-count", help="Number of usable CPUs.", type=int),
            ProviderParameter("cpu-percent", help="Usable percentage of the available CPUs.", type=int),
            ProviderParameter("cpu-period", help="The length of a CPU period in microseconds.", type=int),
            ProviderParameter(
                "cpu-quota", help="Microseconds of CPU time that the container can get in a CPU period.", type=int
            ),
            ProviderParameter("cpu-rt-period", help="Limit CPU real-time period in microseconds.", type=int),
            ProviderParameter("cpu-rt-runtime", help="Limit CPU real-time runtime in microseconds.", type=int),
            ProviderParameter("cpu-shares", help="CPU shares (relative weight).", type=int),
            ProviderParameter("cpuset-cpus", help="CPUs in which to allow execution ('0-3', '0,1').", type=str),
            ProviderParameter(
                "cpuset-mems", help="Memory nodes (MEMs) in which to allow execution ('0-3', '0,1').", type=str
            ),
            ProviderParameter(
                "device-cgroup-rules",
                help="A list of cgroup rules to apply to the container.",
                multiple=True,
                flags=["device-cgroup-rule"],
            ),
            ProviderParameter("devices", help="Expose host devices to the container.", multiple=True, flags=["device"]),
            ProviderParameter("dns", help="Set custom DNS servers.", multiple=True),
            ProviderParameter(
                "dns-opt",
                help="Additional options to be added to the container's ``resolv.conf`` file.",
                type=str,
                flags=["dns-opt", "dns-option"],
            ),
            ProviderParameter("dns-search", help="DNS search domains.", multiple=True),
            ProviderParameter("domainname", help="Container NIS domain name.", type=str),
            ProviderParameter("entrypoint", help="The entrypoint for the container.", type=str),
            ProviderParameter(
                "environment",
                help="Environment variables to set inside the container, in the format 'VAR=VAL'",
                multiple=True,
                flags=["env"],
            ),
            ProviderParameter(
                "group-add",
                help="List of additional group names and/or IDs that the container process will run as.",
                multiple=True,
            ),
            ProviderParameter("hostname", help="Optional hostname for the container.", type=str),
            ProviderParameter(
                "init", help="Run an init inside the container that forwards signals and reaps processes", is_flag=True
            ),
            ProviderParameter("isolation", help="Isolation technology to use.", type=str),
            ProviderParameter("kernel-memory", help="Kernel memory limit (bytes).", type=int, metavar="<bytes>"),
            ProviderParameter("mac-address", help="MAC address to assign to the container.", type=str),
            ProviderParameter("mem-reservation", help="Memory soft limit.", type=int, flags=["memory-reservation"]),
            ProviderParameter(
                "mem-swappiness",
                help="Tune container memory swappiness (0 to 100).",
                type=int,
                flags=["memory-swappiness"],
            ),
            ProviderParameter("memswap-limit", help="Swap limit equal to memory plus swap.", flags=["memory-swap"]),
            ProviderParameter("name", help="The name for this container.", type=str),
            ProviderParameter("network", help="Connect a container to a network.", type=str),
            ProviderParameter("oom-kill-disable", help="Disable OOM Killer.", is_flag=True),
            ProviderParameter("oom-score-adj", help="Tune host's OOM preferences (-1000 to 1000).", type=int),
            ProviderParameter("pids-limit", help="Tune a container's PIDs limit.", type=int),
            ProviderParameter("platform", help="Set platform if server is multi-platform capable.", type=str),
            ProviderParameter("privileged", help="Give extended privileges to this container.", is_flag=True),
            ProviderParameter(
                "publish-all-ports", help="Publish all ports to the host.", is_flag=True, flags=["publish-all"]
            ),
            ProviderParameter("read-only", help="Mount the container's root filesystem as read-only", is_flag=True),
            ProviderParameter("remove", help="Automatically remove the container when it exits.", flags=["rm"]),
            ProviderParameter("runtime", help="Runtime to use with this container.", type=str),
            ProviderParameter("security-opt", help="Security Options.", multiple=True),
            ProviderParameter("shm-size", help="Size of /dev/shm (bytes).", type=int, metavar="<bytes>"),
            ProviderParameter(
                "stdin-open", help="Keep STDIN open even if not attached.", is_flag=True, flags=["interactive"]
            ),
            ProviderParameter("stop-signal", help="Signal to stop the container.", type=str),
            ProviderParameter("tty", help="Allocate a pseudo-TTY.", is_flag=True),
            ProviderParameter("user", help="Username or UID", type=str),
            ProviderParameter("volume-driver", help="The name of a volume driver/plugin.", type=str),
            ProviderParameter(
                "volumes",
                help="A list of volume mounts (e.g. '/host/path/:/mount/path/in/container')",
                multiple=True,
                flags=["volume"],
            ),
            ProviderParameter("volumes-from", help="Mount volumes from the specified container(s)", multiple=True),
        ]

    def get_open_parameters(self) -> List["ProviderParameter"]:
        """Returns parameters that can be set for session open."""
        return []

    def session_list(self, project_name: str) -> List[Session]:
        """Lists all the sessions currently running by the given session provider.

        Returns:
            list: a list of sessions.
        """

        return list(
            map(
                lambda x: Session(
                    id=cast(str, x.short_id),
                    status=x.status,
                    start_time=datetime.fromisoformat(x.labels["start_time"]),
                    provider="docker",
                    commit=x.labels["commit"],
                    branch=x.labels["branch"],
                    url=next(iter(DockerSessionProvider()._get_jupyter_urls(x.ports, x.labels["jupyter_token"]))),
                ),
                self._get_docker_containers(project_name),
            )
        )

    def session_start(
        self,
        image_name: str,
        project_name: str,
        config: Optional[Dict[str, Any]],
        cpu_request: Optional[float] = None,
        mem_request: Optional[str] = None,
        disk_request: Optional[str] = None,
        gpu_request: Optional[Union[str, int]] = None,
        **kwargs,
    ) -> Tuple[str, str]:
        """Creates an interactive session.

        Returns:
            Tuple[str, str]: Provider message and a possible warning message.
        """
        show_non_standard_user_warning = True

        def session_start_helper(consider_disk_request: bool):
            nonlocal gpu_request

            try:
                docker_is_running = self.docker_client().ping()
                if not docker_is_running:
                    raise errors.DockerError(
                        "Could not communicate with the docker instance. Please make sure it is running!"
                    )

                auth_token = uuid4().hex
                default_url = get_value("interactive", "default_url")

                # resource requests
                resource_requests: Dict[str, Any] = dict()
                if cpu_request:
                    # based on the docker go cli: func ParseCPUs
                    resource_requests["nano_cpus"] = int(cpu_request * 10**9)

                if mem_request:
                    resource_requests["mem_limit"] = mem_request

                if consider_disk_request and disk_request:
                    resource_requests["storage_opt"] = {"size": disk_request}

                if gpu_request:
                    if gpu_request == "all":
                        resource_requests["device_requests"] = [
                            docker.types.DeviceRequest(count=-1, capabilities=[["compute", "utility"]])
                        ]
                    else:
                        if not isinstance(gpu_request, int):
                            try:
                                gpu_request = int(gpu_request)
                            except ValueError:
                                raise errors.ParameterError(
                                    f"Invalid value for 'gpu': '{gpu_request}'. Valid values are integers or 'all'"
                                )
                        resource_requests["device_requests"] = [
                            docker.types.DeviceRequest(count=gpu_request, capabilities=[["compute", "utility"]])
                        ]

                # NOTE: set git user
                image_data = self.docker_client().api.inspect_image(image_name)
                working_dir = image_data.get("Config", {}).get("WorkingDir", None)

                if working_dir is None:
                    working_dir = "/home/jovyan"

                work_dir = Path(working_dir) / "work" / project_name.split("/")[-1]

                volumes = kwargs.pop("volumes", [])
                volumes = list(volumes)
                volumes.append(f"{str(project_context.path.resolve())}:{work_dir}")

                environment = {}
                passed_env_vars = kwargs.pop("environment", [])
                for env_var in passed_env_vars:
                    var, _, value = env_var.partition("=")
                    if not var:
                        raise errors.ParameterError(f"Invalid environment variable: '{env_var}'")
                    environment[var] = value

                user = project_context.repository.get_user()
                environment.update(
                    {
                        "GIT_AUTHOR_NAME": user.name,
                        "GIT_AUTHOR_EMAIL": user.email,
                        "GIT_COMMITTER_EMAIL": user.email,
                        "EMAIL": user.email,
                    }
                )

                additional_options: Dict[str, Any] = {}

                if platform.system() == "Linux" and os.getuid() != 1000:
                    # NOTE: Current user id is not 1000 like jovyan, need to run docker under that user.
                    nonlocal show_non_standard_user_warning
                    if show_non_standard_user_warning:
                        communication.confirm(
                            "Your user id is not 1000 and for Jupyter to work the session must be started as root.\n"
                            "Jupyter itself will run as your user.\n"
                            "Starting as root has security implications, make sure you trust this Dockerfile.\n"
                            "Proceed?",
                            abort=True,
                        )
                        show_non_standard_user_warning = False

                    additional_options["user"] = kwargs.pop("user", "root")
                    environment["NB_UID"] = str(os.getuid())
                    environment["CHOWN_HOME"] = "yes"
                    environment["CHOWN_HOME_OPTS"] = "-R"

                container = self.docker_client().containers.run(
                    image_name,
                    'jupyter notebook --NotebookApp.ip="0.0.0.0"'
                    f" --NotebookApp.port={DockerSessionProvider.JUPYTER_PORT}"
                    f' --NotebookApp.token="{auth_token}" --NotebookApp.default_url="{default_url}"'
                    f" --NotebookApp.notebook_dir={work_dir}" + (" --allow-root" if os.getuid() != 1000 else ""),
                    detach=True,
                    labels={
                        "renku_project": project_name,
                        "jupyter_token": auth_token,
                        "commit": project_context.repository.head.commit.hexsha,
                        "branch": project_context.repository.active_branch.name
                        if project_context.repository.active_branch
                        else "master",
                        "start_time": datetime.now().isoformat(),
                    },
                    ports={f"{DockerSessionProvider.JUPYTER_PORT}/tcp": kwargs.get("port")},
                    remove=True,
                    environment=environment,
                    volumes=volumes,
                    working_dir=str(work_dir),
                    **resource_requests,
                    **additional_options,
                    **kwargs,
                )

                if not container.ports:
                    container.reload()

                jupyter_urls = DockerSessionProvider._get_jupyter_urls(
                    container.ports, auth_token, jupyter_port=DockerSessionProvider.JUPYTER_PORT
                )

                message = f"The session for '{image_name}' has been successfully started. It is available at:\n\t"
                message += "\n\t".join(jupyter_urls)
                return message
            except docker.errors.BuildError as error:
                raise errors.DockerError("Couldn't build the image. See inner exception for details.") from error
            except docker.errors.APIError as error:
                raise errors.DockerAPIError("Docker API returned an error. See inner exception for details.") from error
            except ReadTimeout as error:
                raise errors.DockerError(
                    "Couldn't reach the Docker API. Is the docker service running and up to date?"
                ) from error

        try:
            result = session_start_helper(consider_disk_request=True)
        except errors.DockerAPIError:
            warning_message = "Cannot start a session with the disk request: Ignoring the disk request"
            return session_start_helper(consider_disk_request=False), warning_message
        else:
            return result, ""

    def session_stop(self, project_name: str, session_name: Optional[str], stop_all: bool) -> SessionStopStatus:
        """Stops all or a given interactive session."""
        try:
            docker_containers = (
                self._get_docker_containers(project_name)
                if stop_all
                else self.docker_client().containers.list(filters={"id": session_name})
                if session_name
                else self.docker_client().containers.list()
            )

            n_docker_containers = len(docker_containers)

            if n_docker_containers == 0:
                return SessionStopStatus.FAILED if session_name else SessionStopStatus.NO_ACTIVE_SESSION
            elif not session_name and len(docker_containers) > 1:
                return SessionStopStatus.NAME_NEEDED

            [c.stop() for c in docker_containers]
        except docker.errors.APIError as error:
            raise errors.DockerError(error.msg)
        else:
            return SessionStopStatus.SUCCESSFUL

    def session_open(self, project_name: str, session_name: Optional[str], **kwargs) -> bool:
        """Open a given interactive session.

        Args:
            project_name(str): Renku project name.
            session_name(Optional[str]): The unique id of the interactive session.
        """
        url = self.session_url(session_name)

        if not url:
            return False

        webbrowser.open(url)
        return True

    def session_url(self, session_name: Optional[str]) -> Optional[str]:
        """Get the URL of the interactive session."""
        sessions = self.docker_client().containers.list()

        for c in sessions:
            if (
                c.short_id == session_name or (not session_name and len(sessions) == 1)
            ) and f"{DockerSessionProvider.JUPYTER_PORT}/tcp" in c.ports:
                host = c.ports[f"{DockerSessionProvider.JUPYTER_PORT}/tcp"][0]
                return f'http://{host["HostIp"]}:{host["HostPort"]}/?token={c.labels["jupyter_token"]}'
        return None

    def force_build_image(self, force_build: bool = False, **kwargs) -> bool:
        """Whether we should force build the image directly or check for an existing image first."""
        return force_build



#
# Copyright 2018-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Show status of data files created in the repository.

Description
~~~~~~~~~~~

Check if there are output files in a repository that are outdated and need
to be re-generated. Output files get outdated due to changes in input data
or source code (i.e. dependencies).

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.status:status
   :prog: renku status
   :nested: full

Examples
~~~~~~~~

This command shows a list of output files that need to be updated along with
a list of modified inputs for each file. It also display deleted inputs files
if any.

To check for a specific input or output files, you can pass them to this command:

.. code-block:: console

    $ renku status path/to/file1 path/to/file2

In this case, renku only checks if the specified path or paths are modified or
outdated and need an update, instead of checking all inputs and outputs.

The paths mentioned in the output are made relative to the current directory
if you are working in a subdirectory (this is on purpose, to help
cutting and pasting to other commands).

.. cheatsheet::
   :group: Running
   :command: $ renku status
   :description: The the status of generated output files in the project.
   :target: rp
"""

import click

import renku.ui.cli.utils.color as color
from renku.ui.cli.utils.callback import ClickCallback


@click.command()
@click.pass_context
@click.option("-i", "--ignore-deleted", is_flag=True, help="Ignore deleted paths.")
@click.argument("paths", type=click.Path(exists=True, dir_okay=False), nargs=-1)
def status(ctx, paths, ignore_deleted):
    """Show a status of the repository."""
    from renku.command.status import get_status_command

    communicator = ClickCallback()
    result = (
        get_status_command().with_communicator(communicator).build().execute(paths=paths, ignore_deleted=ignore_deleted)
    ).output

    if (
        not result.outdated_outputs
        and not result.deleted_inputs
        and not result.outdated_activities
        and not result.modified_hidden_inputs
    ):
        click.secho("Everything is up-to-date.", fg=color.GREEN)
        return

    outdated = False

    if result.outdated_outputs:
        outdated = True
        click.echo(
            f"Outdated outputs({len(result.outdated_outputs)}):\n"
            "  (use `renku workflow visualize [<file>...]` to see the full lineage)\n"
            "  (use `renku update --all` to generate the file from its latest inputs)\n"
        )
        for k in sorted(result.outdated_outputs.keys()):
            v = result.outdated_outputs[k]
            paths = click.style(", ".join(sorted(v)), fg=color.BLUE, bold=True)
            output = click.style(k, fg=color.RED, bold=True)
            click.echo(f"\t{output}: {paths}")

    if result.modified_hidden_inputs:
        outdated = True
        click.echo(
            f"Outdated workflow files and their outputs({len(result.modified_hidden_inputs)}):\n"
            "  (use `renku run [<workflow-file>]` to generate the outputs from the latest workflow file)\n"
        )
        for k in sorted(result.modified_hidden_inputs.keys()):
            v = result.modified_hidden_inputs[k]
            outputs = click.style(", ".join(sorted(v)), fg=color.RED, bold=True)
            wff = click.style(k, fg=color.BLUE, bold=True)
            click.echo(f"\t{wff}: {outputs}")

        click.echo()

    if not outdated:
        click.secho("All files are generated from the latest inputs.", fg=color.GREEN)

    click.echo()

    if result.modified_inputs:
        click.echo(
            f"Modified inputs({len(result.modified_inputs)}):\n"
            # TODO: Enable once renku workflow visualize is implemented
            # "  (use `renku workflow visualize [<file>...]` to see the full lineage)\n"
        )
        for v in sorted(result.modified_inputs):
            click.echo(click.style(f"\t{v}", fg=color.BLUE, bold=True))
        click.echo()

    if result.deleted_inputs:
        click.echo("Deleted files used to generate outputs:\n")
        for v in sorted(result.deleted_inputs):
            click.echo(click.style(f"\t{v}", fg=color.BLUE, bold=True))
        click.echo()

    if result.outdated_activities:
        click.echo(f"Outdated activities that have no outputs({len(result.outdated_activities)}):\n")
        for k in sorted(result.outdated_activities.keys()):
            v = result.outdated_activities[k]
            paths = click.style(", ".join(sorted(v)), fg=color.BLUE, bold=True)
            activity = click.style(k, fg=color.RED, bold=True)
            click.echo(f"\t{activity}: {paths}")
        click.echo()

    ctx.exit(1 if result.outdated_outputs or result.outdated_activities else 0)



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Plan.

``Plan`` and ``CompositePlan`` classes represent Renku workflow plans executed
in a Project. Each of these classes has a static ``list`` method that returns a
list of all active plans/composite-plans in a project:

.. code-block:: python

    from renku.api import Plan

    plans = Plan.list()

    composite_plans = CompositePlan.list()

"""

from datetime import datetime
from typing import List, Optional, Type, Union, cast

from renku.core import errors
from renku.domain_model.workflow import composite_plan as core_composite_plan
from renku.domain_model.workflow import plan as core_plan
from renku.ui.api.models.parameter import Input, Link, Mapping, Output, Parameter
from renku.ui.api.util import get_plan_gateway


class Plan:
    """API Plan."""

    def __init__(
        self,
        *,
        command: str,
        date_created: Optional[datetime] = None,
        deleted: bool = False,
        description: Optional[str] = None,
        id: str,
        inputs: Optional[List[Input]] = None,
        keywords: Optional[List[str]] = None,
        name: Optional[str] = None,
        outputs: Optional[List[Output]] = None,
        parameters: Optional[List[Parameter]] = None,
        success_codes: Optional[List[int]] = None,
    ):
        self.command: str = command
        self.date_created: Optional[datetime] = date_created
        self.deleted: bool = deleted
        self.description: Optional[str] = description
        self.id: str = id
        self.inputs: List[Input] = inputs or []
        self.keywords: List[str] = keywords or []
        self.name: Optional[str] = name
        self.outputs: List[Output] = outputs or []
        self.parameters: List[Parameter] = parameters or []
        self.success_codes: List[int] = success_codes or []

    @classmethod
    def from_plan(cls, plan: core_plan.Plan) -> "Plan":
        """Create an instance from a core Plan model.

        Args:
            plan(core_plan.Plan): The core plan.

        Returns:
            Plan: An API Plan model.
        """
        return cls(
            command=plan.command,
            date_created=plan.date_created,
            deleted=plan.date_removed is not None,
            description=plan.description,
            id=plan.id,
            inputs=[Input.from_parameter(i) for i in plan.inputs],
            keywords=plan.keywords,
            name=plan.name,
            outputs=[Output.from_parameter(o) for o in plan.outputs],
            parameters=[Parameter.from_parameter(p) for p in plan.parameters],
            success_codes=plan.success_codes,
        )

    @staticmethod
    def list(include_deleted: bool = False) -> List[Union["Plan", "CompositePlan"]]:
        """List all plans in a project.

        Args:
            include_deleted(bool): Whether to include deleted plans.

        Returns:
            A list of all plans in the supplied project.
        """
        return _convert_plans(_list_plans(include_deleted=include_deleted, type=core_plan.Plan))

    def __repr__(self):
        return f"<Plan '{self.name}'>"

    @property
    def activities(self) -> List:
        """Return a list of upstream activities."""
        from renku.ui.api.models.activity import get_activities

        return get_activities(plan_id=self.id)

    def get_latest_version(self) -> "Plan":
        """Return the latest version (derivative) of this plan."""
        return cast(Plan, _get_latest_version(plan=self))


class CompositePlan:
    """API CompositePlan."""

    def __init__(
        self,
        *,
        date_created: Optional[datetime] = None,
        deleted: bool = False,
        description: Optional[str] = None,
        id: str,
        keywords: Optional[List[str]] = None,
        links: Optional[List[Link]] = None,
        mappings: Optional[List[Mapping]] = None,
        name: Optional[str] = None,
        plans: Optional[List[Union["CompositePlan", "Plan"]]] = None,
    ):
        self.date_created: Optional[datetime] = date_created
        self.deleted: bool = deleted
        self.description: Optional[str] = description
        self.id: str = id
        self.keywords: List[str] = keywords or []
        self.links: List[Link] = links or []
        self.mappings: List[Mapping] = mappings or []
        self.name: Optional[str] = name
        self.plans: List[Union["CompositePlan", "Plan"]] = plans or []

    @classmethod
    def from_composite_plan(cls, composite_plan: core_composite_plan.CompositePlan) -> "CompositePlan":
        """Create an instance from a core CompositePlan model.

        Args:
            composite_plan(core_composite_plan.CompositePlan): The core composite plan.

        Returns:
            CompositePlan: An API CompositePlan instance.
        """
        return cls(
            date_created=composite_plan.date_created,
            deleted=composite_plan.date_removed is not None,
            description=composite_plan.description,
            id=composite_plan.id,
            keywords=composite_plan.keywords,
            links=[Link.from_link(link) for link in composite_plan.links],
            mappings=[Mapping.from_parameter(m) for m in composite_plan.mappings],
            name=composite_plan.name,
            plans=_convert_plans(composite_plan.plans),
        )

    @staticmethod
    def list(include_deleted: bool = False) -> List[Union["Plan", "CompositePlan"]]:
        """List all plans in a project.

        Args:
            include_deleted(bool): Whether to include deleted plans.

        Returns:
            A list of all plans in the supplied project.
        """
        return _convert_plans(_list_plans(include_deleted=include_deleted, type=core_composite_plan.CompositePlan))

    def __repr__(self):
        return f"<CompositePlan '{self.name}'>"

    @property
    def activities(self) -> List:
        """Return a list of upstream activities."""
        from renku.ui.api.models.activity import get_activities

        return get_activities(plan_id=self.id)

    def get_latest_version(self) -> "CompositePlan":
        """Return the latest version (derivative) of this plan."""
        return cast(CompositePlan, _get_latest_version(plan=self))


def _convert_plan(plan) -> Union[Plan, CompositePlan]:
    """Convert a core Plans/CompositePlans to API Plans/CompositePlans."""
    if isinstance(plan, core_plan.Plan):
        return Plan.from_plan(plan)
    elif isinstance(plan, core_composite_plan.CompositePlan):
        return CompositePlan.from_composite_plan(plan)

    raise errors.ParameterError(f"Invalid plan type: {type(plan)}")


def _convert_plans(plans: List[Union[core_plan.AbstractPlan]]) -> List[Union[Plan, CompositePlan]]:
    """Convert a list of core Plans/CompositePlans to API Plans/CompositePlans."""
    return [_convert_plan(p) for p in plans]


def _list_plans(
    include_deleted: bool, type: Optional[Type[Union[core_plan.Plan, core_composite_plan.CompositePlan]]]
) -> List[core_plan.AbstractPlan]:
    """List all plans in a project.

    Args:
        include_deleted(bool): Whether to include deleted plans.
        type(Optional[Type[Union[core_plan.Plan, core_composite_plan.CompositePlan]]]): The type of plan to list.
    Returns:
        A list of all plans in the supplied project.
    """
    plan_gateway = get_plan_gateway()
    if not plan_gateway:
        return []

    plans = plan_gateway.get_all_plans()
    derivatives_mapping = {p.derived_from: p for p in plans if p.derived_from is not None}

    def get_latest_plan(plan):
        while plan.id in derivatives_mapping:
            plan = derivatives_mapping[plan.id]
        return plan

    latest_plans = {get_latest_plan(p) for p in plans}

    if not include_deleted:
        latest_plans = {p for p in latest_plans if not p.deleted}

    return [p for p in latest_plans if type is None or isinstance(p, type)]


def _get_latest_version(plan: Union[Plan, CompositePlan]) -> Union[Plan, CompositePlan]:
    """Get the latest version (derivative) of a plan or the plan itself if it's already the latest version."""
    all_plans = _list_plans(include_deleted=False, type=None)
    derivatives_mapping = {p.derived_from: p for p in all_plans if p.derived_from is not None}

    derived_plan = None
    plan_id = plan.id
    while plan_id in derivatives_mapping:
        derived_plan = derivatives_mapping[plan_id]
        plan_id = derived_plan.id

    return _convert_plan(derived_plan) if derived_plan is not None else plan



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Activity graph view model."""

from datetime import datetime
from itertools import repeat
from textwrap import shorten
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple

from renku.core import errors

if TYPE_CHECKING:
    from grandalf.graphs import Edge

    from renku.command.view_model.text_canvas import Point, TextCanvas

MAX_NODE_LENGTH = 40


class NodeViewer:
    """Gets width/height for a display node."""

    def __init__(self, text):
        lines = text.splitlines()
        self.w = max([len(line) for line in lines]) + 2
        self.h = len(lines) + 2


class ActivityGraphViewModel:
    """A view model for outputting a directed activity graph."""

    def __init__(self, graph):
        self.graph = graph

    def _format_vertex(self, node, columns: List[Callable]) -> Tuple[str, bool, Any]:
        """Return vertex text for a node.

        Args:
            node: The node to format.
            columns (List[str]): The fields to include in the node text.

        Returns:
            Tuple[str, bool, Any]: Tuple of string representation of node, whether or not
                it is an Activity and the Activity if it is.
        """
        from renku.domain_model.provenance.activity import Activity

        if isinstance(node, Activity):
            return "\n".join(shorten(c(node), MAX_NODE_LENGTH, placeholder="...") for c in columns), True, node

        return node, False, None

    def _subgraph_order_key(self, subgraph) -> datetime:
        """Return a sorting key for ordering subgraphs.

        Subgraphs are ordered by the ended_at_time of the lastest activities in them.

        Args:
            subgraph: Subgraph to get an order key for.

        Returns:
            datetime: The latest activity time in the subgraph.
        """
        from renku.domain_model.provenance.activity import Activity

        activity_times = [n.ended_at_time for n in subgraph.nodes() if isinstance(n, Activity)]

        return max(activity_times)

    def _format_vertex_raw(self, node, columns: List[Callable]) -> str:
        """Return vertex text for a node.

        Args:
            node: The node to format.
            columns (List[Callable]): The fields to include in the node text.

        Returns:
            string representation of node
        """
        import json

        from renku.domain_model.provenance.activity import Activity

        if isinstance(node, Activity):
            text = "\n".join(c(node) for c in columns)
        else:
            text = node

        # NOTE: double quotes are common in console command, repr() wouldn't escape properly
        return json.dumps(text)

    def _get_lambda_columns(self, columns):
        """Return lambda columns.

        Args:
            columns (str): comma-separated column names.

        Returns:
            List[Callable] lambda columns
        """

        try:
            return [ACTIVITY_GRAPH_COLUMNS[c] for c in columns.split(",")]
        except KeyError as e:
            wrong_values = ", ".join(e.args)
            suggestion = ",".join(ACTIVITY_GRAPH_COLUMNS.keys())
            raise errors.ParameterError(f"you can use any of '{suggestion}'", f"columns '{wrong_values}'")

    def layout_graph(self, columns):
        """Create a Sugiyama layout of the graph.

        Args:
            columns: Columns to show as node text.
        """
        import networkx
        from grandalf.graphs import Edge, Graph, Vertex
        from grandalf.layouts import SugiyamaLayout
        from grandalf.routing import EdgeViewer, route_with_lines

        from renku.domain_model.provenance.activity import Activity

        columns = self._get_lambda_columns(columns)

        self.layouts: List[SugiyamaLayout] = []

        components = networkx.weakly_connected_components(self.graph)
        subgraphs = [self.graph.subgraph(component).copy() for component in components]
        subgraphs = list(filter(lambda s: any(isinstance(n, Activity) for n in s), subgraphs))
        subgraphs = sorted(subgraphs, key=self._subgraph_order_key)

        for subgraph in subgraphs:
            nodes = {n: Vertex(self._format_vertex(n, columns)) for n in subgraph.nodes}
            edges = [Edge(nodes[s], nodes[t]) for s, t in subgraph.edges]
            roots = [nodes[n] for n, d in subgraph.in_degree() if d == 0]

            graph = Graph(nodes.values(), edges)

            shortest_node_width = 2**64
            shortest_node_height = 2**64

            for node in nodes.values():
                node.view = NodeViewer(node.data[0])
                shortest_node_width = min(shortest_node_width, node.view.w)
                shortest_node_height = min(shortest_node_height, node.view.h)

            for edge in edges:
                edge.view = EdgeViewer()

            layout = SugiyamaLayout(graph.C[0])
            layout.init_all(roots=roots, optimize=True)
            layout.xspace = max(shortest_node_width / 2, 20)
            layout.yspace = shortest_node_height

            # space between beginning of one node and beginning of next
            self.vertical_space = shortest_node_height * 2

            layout.route_edge = route_with_lines
            layout.draw(5)
            self.layouts.append(layout)

    def _add_edges_to_canvas(
        self, edges: List["Edge"], canvas: "TextCanvas", existing_edges: List["Edge"], min_y
    ) -> Tuple[int, str]:
        """Add an edge to a canvas object.

        Makes sure overlapping edges don't have the same color.

        Args:
            edges(List["Edge"]): Sub-edges of the edge to add.
            canvas("TextCanvas"): Canvas to add edge to.
            existing_edges(List["Edge"]): Already existing edges on canvas.
            min_y: Starting y coordinate inside canvas.

        Returns:
            Tuple[int, str]: Ending y coordinate after edge is added and color used for edge.
        """
        from renku.command.view_model.text_canvas import EdgeShape, Point

        edge_color = EdgeShape.next_color()
        new_edges = []
        max_y = 0

        for edge in edges:
            points = edge.view._pts
            for index in range(len(points) - 1):
                start = points[index]
                end = points[index + 1]
                max_y = max(max_y, end[1])
                new_edges.append(
                    EdgeShape(Point(start[0], start[1] + min_y), Point(end[0], end[1] + min_y), color=edge_color)
                )

        # figure out if this edge crosses any existing edge with the same color, if so, change the color
        same_color_edges = [e for e in existing_edges if e.color == edge_color]

        if same_color_edges:
            # check for intersections
            intersecting_edges = [e for e in same_color_edges if any(e.intersects_with(o) for o in new_edges)]

            colors = {e.color for e in intersecting_edges}

            if len(colors) < len(EdgeShape.COLORS):
                while edge_color in colors:
                    edge_color = EdgeShape.next_color()
                for e in new_edges:
                    e.color = edge_color

        for e in new_edges:
            canvas.add_shape(e, layer=0)

        existing_edges.extend(new_edges)
        return max_y, edge_color

    def dot_representation(self, columns: str) -> str:
        """Return the graph as a Graphviz Dot string.

        Args:
            columns(str): Columns to include in node text.

        Returns:
            string representing the Graphviz Dot graph
        """
        import io

        from renku.domain_model.provenance.activity import Activity

        # compute node text
        columns_callable = self._get_lambda_columns(columns)
        activities_text = {}
        for node in self.graph.nodes:
            if isinstance(node, Activity):
                output_text = "\n".join(c(node) for c in columns_callable)
                activities_text[str(node)] = output_text

        output = io.StringIO()
        output.write("digraph {\n")

        # add edges and track visited nodes
        visited_nodes = []
        for edge in self.graph.edges:
            vertexes = tuple(map(self._format_vertex_raw, edge, repeat(columns_callable, 2)))
            output.write(f"{vertexes[0]} -> {vertexes[1]};")
            for vertex in vertexes:
                if vertex not in visited_nodes:
                    visited_nodes.append(vertex)

        # add missing nodes
        for node in self.graph.nodes:
            lonely_node = self._format_vertex_raw(node, columns_callable)
            if lonely_node not in visited_nodes:
                output.write(f'"{lonely_node}";')

        output.write("\n}")
        return output.getvalue()

    def text_representation(
        self, columns: str, color: bool = True, ascii=False
    ) -> Tuple[Optional[str], Optional[List[List[Tuple["Point", "Point", Any]]]]]:
        """Return an ascii representation of the graph.

        Args:
            columns(str): Columns to include in node text.
            color(bool, optional): Whether to render in color or not (Default value = True).
            ascii: Whether to use only ascii characters or also UTF8  (Default value = False).

        Returns:
            Tuple[str, List[List[Tuple[Point, Point, Any]]]]: Tuple of rendered canvas text and coordinates
                of nodes within canvas.
        """
        from grandalf.layouts import DummyVertex

        from renku.command.view_model.text_canvas import NodeShape, Point, TextCanvas
        from renku.domain_model.provenance.activity import Activity

        self.layout_graph(columns=columns)

        canvas = TextCanvas()
        navigation_data = []

        # min_y sets the minimum y coordinate for each layout(distinct subgraph) in the graph
        min_y = 0

        if not self.layouts:
            # nothing to show
            return None, None

        for layout in self.layouts:
            max_y = 0
            existing_edges: List[Edge] = []

            for layer in layout.layers:
                layer_nodes = []

                for node in layer:
                    if isinstance(node, DummyVertex):
                        continue

                    node_color = None

                    if node.data[2] and isinstance(node.data[2], Activity):
                        # NOTE: get edges for node
                        connected_edges = list(layout.g.E(cond=lambda e: e.v[0] == node or e.v[1] == node))

                        visited_edges = set()
                        visited_nodes = {node}

                        # NOTE: Follow all edges connected to the node (might have DummyNode's in between)
                        while connected_edges:
                            current_edge = connected_edges.pop()
                            visited_edges.add(current_edge)

                            dummy_node = None

                            if isinstance(current_edge.v[0], DummyVertex) and current_edge.v[0] not in visited_nodes:
                                dummy_node = current_edge.v[0]
                            elif isinstance(current_edge.v[1], DummyVertex) and current_edge.v[0] not in visited_nodes:
                                dummy_node = current_edge.v[1]

                            if dummy_node:
                                connected_edges.extend(e for e in dummy_node.e if e not in visited_edges)
                                visited_nodes.add(dummy_node)

                        local_max_y, node_color = self._add_edges_to_canvas(
                            list(visited_edges), canvas, existing_edges, min_y
                        )
                        max_y = max(max_y, local_max_y)

                    xy = node.view.xy
                    split = node.data[0].splitlines()
                    max_height = len(split)
                    max_width = max(len(line) for line in split)
                    node_shape = NodeShape(
                        node.data[0],
                        Point(round(xy[0] - max_width / 2), round(xy[1] + min_y - max_height)),
                        double_border=node.data[1],
                        color=node_color,
                    )
                    canvas.add_shape(node_shape, layer=1)
                    max_y = max(max_y, node_shape.extent[0][1])

                    if node.data[2]:
                        layer_nodes.append((node_shape, node.data[2]))

                if layer_nodes:
                    # add layer for navigation
                    navigation_data.append(layer_nodes)

            # move subsequent layouts down below current layout
            min_y += max_y + 5

        canvas.render(color=color, ascii=ascii)

        # adjust coordinates to actual coordinates in layers
        transformed_navigation_data = []
        for layer in navigation_data:
            node_list = []
            for node in layer:
                top_left, bottom_right = node[0].actual_extent
                start = canvas.get_coordinates(top_left)
                end = canvas.get_coordinates(bottom_right)
                node_list.append((start, end, node[1]))
            transformed_navigation_data.append(node_list)

        return canvas.text, transformed_navigation_data


ACTIVITY_GRAPH_COLUMNS = {
    "command": lambda a: " ".join(a.plan_with_values.to_argv(with_streams=True)),
    "id": lambda a: a.id,
    "date": lambda a: a.ended_at_time.isoformat(),
    "plan": lambda a: a.plan_with_values.name,
}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku save commands."""

from typing import List, Optional, Tuple

from pydantic import validate_arguments

from renku.command.command_builder.command import Command
from renku.core import errors
from renku.core.storage import track_paths_in_storage
from renku.domain_model.project_context import project_context


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _save_and_push(
    message: Optional[str] = None, remote: Optional[str] = None, paths: Optional[List[str]] = None
) -> Tuple[List[str], str]:
    """Save and push local changes.

    Args:
        message(Optional[str]): The commit message (Default value = None).
        remote(Optional[str]): The remote to push to (Default value = None).
        paths(Optional[List[str]]): The paths to include in the commit (Default value = None).

    Returns:
        Tuple[List[str], str]: Tuple of paths that were committed and branch that was pushed.
    """
    from renku.core.util.git import commit_changes, get_dirty_paths, get_remote, push_changes

    repository = project_context.repository

    setup_credential_helper()

    if not paths:
        paths = list(get_dirty_paths(repository))
    else:
        staged_changes = repository.staged_changes
        if staged_changes:
            staged_paths = {c.a_path for c in staged_changes}
            not_passed = staged_paths - set(paths)

            if not_passed:
                raise errors.RenkuSaveError(
                    "These files are in the git staging area, but weren't passed to renku save. Unstage them or pass"
                    + " them explicitly: \n"
                    + "\n".join(not_passed)
                )

    # NOTE: Check if a remote is setup for the repository
    if not remote:
        default_remote = get_remote(repository)
        if not default_remote:
            raise errors.GitError("No remote has been set up for the current branch")

    if paths:
        track_paths_in_storage(*paths)
        paths = commit_changes(*paths, repository=repository, message=message)

    branch = push_changes(repository=repository, remote=remote)

    return paths, branch


def setup_credential_helper():
    """Setup git credential helper to ``cache`` if not set already."""
    repository = project_context.repository

    credential_helper = repository.get_configuration().get_value("credential", "helper", "")

    if not credential_helper:
        with repository.get_configuration(writable=True) as w:
            w.set_value("credential", "helper", "cache")


def save_and_push_command():
    """Command to save and push."""
    return Command().command(_save_and_push)



#
# Copyright 2018-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Clone a Renku repo along with all Renku-specific initializations."""

from pathlib import Path
from typing import Any, Dict, Optional, Union

from git.remote import RemoteProgress
from pydantic import validate_arguments

from renku.command.command_builder.command import Command
from renku.domain_model.project_context import project_context


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _project_clone(
    url: str,
    path: Optional[Union[str, Path]] = None,
    install_githooks: bool = True,
    install_mergetool: bool = True,
    skip_smudge: bool = True,
    recursive: bool = True,
    depth: Optional[int] = None,
    progress: Optional[RemoteProgress] = None,
    config: Optional[Dict[str, Any]] = None,
    raise_git_except: bool = False,
    checkout_revision: Optional[str] = None,
    use_renku_credentials: bool = False,
):
    """Clone Renku project repo, install Git hooks and LFS.

    Args:
        url(str): Git URL to clone.
        path(Optional[str]): Path to clone to (Default value = None).
        install_githooks(bool): Whether to install the pre-commit hook or not (Default value = True).
        install_mergetool(bool): Whether to install the renku metadata git mergetool or not (Default value = True).
        skip_smudge(bool): Whether to skip pulling files from LFS (Default value = True).
        recursive(bool): Recursively clone (Default value = True).
        depth(Optional[int]): Clone depth (commits from HEAD) (Default value = None).
        progress(Optional[RemoteProgress]): Git progress object (Default value = None).
        config(Optional[Dict[str, Any]]): Initial config (Default value = None).
        raise_git_except(bool): Whether to raise Git exceptions or not (Default value = False).
        checkout_revision(Optional[str]): Specific revision to check out (Default value = None).
        use_renku_credentials(bool): Whether to use credentials stored in renku (Default value = False).

    Returns:
        Tuple of cloned ``Repository`` and whether it's a Renku project or not.
    """
    from renku.command.mergetool import setup_mergetool
    from renku.core.util.git import clone_renku_repository
    from renku.core.util.metadata import is_renku_project

    install_lfs = project_context.external_storage_requested

    repository = clone_renku_repository(
        url=url,
        path=path,
        install_githooks=install_githooks,
        install_lfs=install_lfs,
        skip_smudge=skip_smudge,
        recursive=recursive,
        depth=depth,
        progress=progress,
        config=config,
        raise_git_except=raise_git_except,
        checkout_revision=checkout_revision,
        use_renku_credentials=use_renku_credentials,
    )

    with project_context.with_path(repository.path):
        project_initialized = is_renku_project()

        if project_initialized and install_mergetool:
            setup_mergetool(with_attributes=False)

    return repository, project_initialized


def project_clone_command():
    """Command to clone a renku project."""
    return Command().command(_project_clone).with_database()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Job queues."""
import os

from rq import Queue

from renku.ui.service.cache.base import BaseCache
from renku.ui.service.cache.config import REDIS_NAMESPACE

CLEANUP_QUEUE_FILES = f"{REDIS_NAMESPACE}.cache.cleanup.files"
CLEANUP_QUEUE_PROJECTS = f"{REDIS_NAMESPACE}.cache.cleanup.projects"

DATASETS_JOB_QUEUE = f"{REDIS_NAMESPACE}.datasets.jobs"
MIGRATIONS_JOB_QUEUE = f"{REDIS_NAMESPACE}.project.migrations"
GRAPH_JOB_QUEUE = f"{REDIS_NAMESPACE}.graph.jobs"

DELAYED_CTRL_DATASETS_CREATE = f"{REDIS_NAMESPACE}.delayed.ctrl.DatasetsCreateCtrl"
DELAYED_CTRL_DATASETS_ADD = f"{REDIS_NAMESPACE}.delayed.ctrl.DatasetsAddFileCtrl"
DELAYED_CTRL_DATASETS_REMOVE = f"{REDIS_NAMESPACE}.delayed.ctrl.DatasetsRemoveCtrl"
DELAYED_CTRL_DATASETS_IMPORT = f"{REDIS_NAMESPACE}.delayed.ctrl.DatasetsImportCtrl"
DELAYED_CTRL_DATASETS_EDIT = f"{REDIS_NAMESPACE}.delayed.ctrl.DatasetsEditCtrl"
DELAYED_CTRL_DATASETS_UNLINK = f"{REDIS_NAMESPACE}.delayed.ctrl.DatasetsUnlinkCtrl"
DELAYED_CTRL_PROJECT_MIGRATE = f"{REDIS_NAMESPACE}.delayed.ctrl.MigrateProjectCtrl"
DELAYED_CTRL_CONFIG_SET = f"{REDIS_NAMESPACE}.delayed.ctrl.SetConfigCtrl"


QUEUES = [
    CLEANUP_QUEUE_FILES,
    CLEANUP_QUEUE_PROJECTS,
    DATASETS_JOB_QUEUE,
    MIGRATIONS_JOB_QUEUE,
    GRAPH_JOB_QUEUE,
    DELAYED_CTRL_DATASETS_CREATE,
    DELAYED_CTRL_DATASETS_ADD,
    DELAYED_CTRL_DATASETS_REMOVE,
    DELAYED_CTRL_DATASETS_IMPORT,
    DELAYED_CTRL_DATASETS_EDIT,
    DELAYED_CTRL_DATASETS_UNLINK,
    DELAYED_CTRL_PROJECT_MIGRATE,
    DELAYED_CTRL_CONFIG_SET,
]


class WorkerQueues:
    """Worker queues."""

    connection = BaseCache.cache

    @staticmethod
    def describe():
        """List possible queues."""
        return QUEUES

    @staticmethod
    def get(name):
        """Get specific queue object."""
        return Queue(
            name, connection=WorkerQueues.connection, default_timeout=int(os.getenv("WORKER_DEFAULT_JOBS_TIMEOUT", 600))
        )



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for renku service controllers."""
import shutil

from renku.command.clone import project_clone_command
from renku.core.util.contexts import renku_project_context
from renku.ui.service.cache.models.project import Project
from renku.ui.service.logger import service_log
from renku.ui.service.views.decorators import requires_cache


@requires_cache
def user_project_clone(cache, user_data, project_data):
    """Clones the project for a given user."""
    if "project_id" in project_data:
        project_data.pop("project_id")

    user = cache.ensure_user(user_data)
    project = cache.make_project(user, project_data, persist=False)

    # NOTE: Create parent dir so lock file can be created.
    project.abs_path.parent.mkdir(parents=True, exist_ok=True)

    with project.write_lock(), renku_project_context(project.abs_path, check_git_path=False):
        git_url = project_data.get("git_url")

        if git_url is not None:
            try:
                # NOTE: If two requests ran at the same time, by the time we acquire the lock a project might already
                # be cloned by an earlier request.
                found_project = Project.get(
                    (Project.user_id == user_data["user_id"])
                    & (Project.git_url == git_url)
                    & (Project.project_id != project.project_id)
                )
            except ValueError:
                pass
            else:
                service_log.debug(f"project already cloned, skipping clone: {git_url}")
                return found_project

        if project.abs_path.exists():
            # NOTE: Remove dir since a previous clone might have failed somewhere in the middle.
            shutil.rmtree(str(project.abs_path))

        project.abs_path.mkdir(parents=True, exist_ok=True)

        repo, project.initialized = (
            project_clone_command()
            .build()
            .execute(
                project_data["url_with_auth"],
                path=project.abs_path,
                depth=project_data["depth"],
                raise_git_except=True,
                config={
                    "user.name": project_data["fullname"],
                    "user.email": project_data["email"],
                    "pull.rebase": False,
                },
                checkout_revision=project_data["branch"],
            )
        ).output
        project.save()

    service_log.debug(f"project successfully cloned: {repo}")

    return project



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for converting metadata."""

from pathlib import Path
from typing import List, Optional, Tuple, Union
from urllib.parse import urlparse

from renku.core.migration.models import v9 as old_datasets
from renku.core.util.datetime8601 import fix_datetime
from renku.core.util.git import get_entity_from_revision
from renku.core.util.urls import get_slug
from renku.domain_model.dataset import (
    Dataset,
    DatasetFile,
    DatasetTag,
    ImageObject,
    Language,
    RemoteEntity,
    Url,
    is_dataset_name_valid,
)
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance import agent as new_agents


def _convert_dataset_identifier(identifier: str) -> str:
    """Remove - from a dataset identifier."""
    return identifier.replace("-", "")


def _convert_dataset_tag(tag: Optional[old_datasets.DatasetTag]) -> Optional[DatasetTag]:
    """Convert old DatasetTag to new DatasetTag."""
    if not tag:
        return

    # NOTE: ``dataset_id`` field will be set later when processing the migrated commit.
    return DatasetTag(
        dataset_id=Url(url_id="dummy-id"), date_created=tag.created, description=tag.description, name=tag.name
    )


def _convert_language(language: Optional[old_datasets.Language]) -> Optional[Language]:
    """Convert old Language to new Language."""
    if not language:
        return
    return Language(name=language.name, alternate_name=language.alternate_name)


def _convert_image_object(image_object: Optional[old_datasets.ImageObject], dataset_id: str) -> Optional[ImageObject]:
    """Create from old ImageObject instance."""
    if not image_object:
        return
    id = ImageObject.generate_id(dataset_id=dataset_id, position=image_object.position)
    return ImageObject(content_url=image_object.content_url, position=image_object.position, id=id)


def _create_remote_entity(dataset_file: Optional[old_datasets.DatasetFile]) -> Optional[RemoteEntity]:
    """Create RemoteEntity from old DatasetFile."""
    if not dataset_file:
        return
    commit_sha = dataset_file._label.rsplit("@", maxsplit=1)[-1]
    return RemoteEntity(checksum=commit_sha, path=dataset_file.path, url=dataset_file.url)


def _convert_dataset_file(dataset_file: old_datasets.DatasetFile, revision: str) -> Optional[DatasetFile]:
    """Convert old DatasetFile to new DatasetFile if available at revision."""
    repository = project_context.repository

    entity = get_entity_from_revision(repository=repository, path=dataset_file.path, revision=revision)
    if not entity:
        return

    return DatasetFile(
        based_on=_create_remote_entity(dataset_file.based_on),
        date_added=dataset_file.added,
        entity=entity,
        is_external=dataset_file.external,
        source=dataset_file.source,
    )


def _convert_person(person: Optional[old_datasets.Person]) -> Optional[new_agents.Person]:
    """Create a Person from and old Person."""
    if not person:
        return

    return new_agents.Person(
        affiliation=person.affiliation,
        alternate_name=person.alternate_name,
        email=person.email,
        id=None,
        name=person.name,
    )


def _convert_agent(
    agent: Optional[Union[old_datasets.Person, old_datasets.SoftwareAgent]]
) -> Optional[Union[new_agents.Person, new_agents.SoftwareAgent]]:
    """Create an instance from Person/SoftwareAgent."""
    if isinstance(agent, old_datasets.SoftwareAgent):
        return new_agents.SoftwareAgent(id=agent.id, name=agent.label)

    assert not agent or isinstance(agent, old_datasets.Person), f"Invalid type {type(agent)}"
    return _convert_person(agent)


def _convert_same_as(url: Optional[old_datasets.Url]) -> Optional[Url]:
    """Convert old Url to new Url."""
    if not url:
        return

    if url.url_str:
        parsed_url = urlparse(url.url_str)
    else:
        assert url.url_id
        parsed_url = urlparse(url.url_id)

    # NOTE: Test if dataset is imported from a renku deployment. This is good enough for all current deployments.
    if "renku" in parsed_url.netloc:
        path = _convert_dataset_identifier(parsed_url.path)
        parsed_url = parsed_url._replace(path=path)

    url_str = parsed_url.geturl()

    return Url(url_str=url_str) if url.url_str else Url(url_id=url_str)


def convert_dataset(dataset: old_datasets.Dataset, revision: str) -> Tuple[Dataset, List[DatasetTag]]:
    """Convert old Dataset to new Dataset."""

    def convert_dataset_files(files: List[old_datasets.DatasetFile]) -> List[DatasetFile]:
        """Create instances from old DatasetFile."""
        dataset_files = []
        files = {f.path: f for f in files}  # NOTE: To make sure there are no duplicate paths

        for file in files.values():
            new_file = _convert_dataset_file(dataset_file=file, revision=revision)
            if not new_file:
                continue

            dataset_files.append(new_file)

        return dataset_files

    def convert_derived_from(
        derived_from: Optional[old_datasets.Url], same_as: Optional[old_datasets.Url]
    ) -> Optional[Url]:
        """Return Dataset.id from `derived_from` url."""
        if not derived_from or same_as:
            return

        url = derived_from.url.get("@id")
        path = _convert_dataset_identifier(urlparse(url).path)

        return Url(url_id=Dataset.generate_id(identifier=Path(path).name))

    def convert_license(license):
        if not license:
            return license
        elif isinstance(license, (Url, str)):
            return license
        elif isinstance(license, dict) and len(license) == 1:
            return list(license.values())[0]
        elif isinstance(license, dict) and "@id" in license:
            return license["@id"]
        elif isinstance(license, list) and len(license) == 1:
            return license[0]

        return str(license)

    tags = [_convert_dataset_tag(tag) for tag in (dataset.tags or [])]
    name = get_slug(dataset.name) if not is_dataset_name_valid(dataset.name) else dataset.name

    identifier = _convert_dataset_identifier(dataset.identifier)
    id = Dataset.generate_id(identifier=identifier)

    return (
        Dataset(
            creators=[_convert_agent(creator) for creator in dataset.creators],
            dataset_files=convert_dataset_files(dataset.files),
            date_created=fix_datetime(dataset.date_created),
            date_published=fix_datetime(dataset.date_published),
            date_removed=None,
            date_modified=fix_datetime(dataset.date_created or dataset.date_published),
            derived_from=convert_derived_from(dataset.derived_from, dataset.same_as),
            description=dataset.description,
            id=id,
            identifier=identifier,
            images=[_convert_image_object(image, dataset_id=id) for image in (dataset.images or [])],
            in_language=_convert_language(dataset.in_language),
            keywords=dataset.keywords,
            license=convert_license(dataset.license),
            name=name,
            project_id=project_context.project.id,
            initial_identifier=_convert_dataset_identifier(dataset.initial_identifier),
            same_as=_convert_same_as(dataset.same_as),
            title=dataset.title,
            version=dataset.version,
        ),
        tags,
    )



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resolution of ``Workflow`` execution values precedence."""

import datetime
from abc import ABC, abstractmethod
from itertools import chain
from string import Formatter
from typing import Any, Dict, Iterable, Mapping, Optional, Set, Tuple, Union

from renku.core import errors
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.parameter import CommandParameterBase, ParameterMapping
from renku.domain_model.workflow.plan import AbstractPlan, Plan


class TemplateVariableFormatter(Formatter):
    """Template variable formatter for `CommandParameterBase`."""

    RESERVED_KEYS = ["iter_index"]

    def __init__(self):
        super().__init__()

    def apply(self, param: str, parameters: Optional[Mapping[str, Any]] = None) -> str:
        """Renders the parameter template into its final value."""
        if parameters is None:
            parameters = {}

        try:
            return super().vformat(param, args=[datetime.datetime.now()], kwargs=parameters)
        except KeyError as e:
            raise errors.ParameterError(f"Could not resolve the variable {str(e)}")

    def get_value(self, key, args, kwargs):
        """Ignore some special keys when formatting the variable."""
        if key in self.RESERVED_KEYS:
            return key
        return super().get_value(key, args, kwargs)

    @staticmethod
    def to_map(parameters: Iterable[Union[CommandParameterBase, Tuple[str, str]]]) -> Mapping[str, str]:
        """Converts a list of `CommandParameterBase` into parameter name-value dictionary."""
        return dict(
            map(
                lambda x: (x.name, x.actual_value) if isinstance(x, CommandParameterBase) else (x[1], str(x[0])),
                parameters,
            )
        )


class ValueResolver(ABC):
    """Value resolution class for an ``AbstractPlan``."""

    def __init__(self, plan: AbstractPlan, values: Optional[Dict[str, Any]]):
        self._values = values
        self.missing_parameters: Set[str] = set()
        self._plan = plan
        self._template_engine = TemplateVariableFormatter()

    @abstractmethod
    def apply(self) -> AbstractPlan:
        """Applies values and default_values to a potentially nested workflow.

        Returns:
            AbstractPlan: The ``AbstractPlan`` with the user provided values set.
        """
        pass

    @staticmethod
    def get(plan: AbstractPlan, values: Dict[str, Any]) -> "ValueResolver":
        """Factory method to obtain the specific ValueResolver for a workflow.

        Args:
            plan(AbstractPlan): a workflow.
            values(Dict[str, Any]): user defined dictionary of runtime values for the provided workflow.

        Returns:
            "ValueResolver": A ValueResolver object.
        """
        return PlanValueResolver(plan, values) if isinstance(plan, Plan) else CompositePlanValueResolver(plan, values)


class PlanValueResolver(ValueResolver):
    """Value resolution class for a ``Plan``.

    Applies values and default_values to a workflow.
    """

    def __init__(self, plan: Plan, values: Dict[str, Any]):
        super().__init__(plan, values)

    def apply(self) -> AbstractPlan:
        """Applies values and default_values to a ``Plan``.

        Returns:
            A Plan with values applied.
        """
        if self._values is None:
            return self._plan

        values_keys = set(self._values.keys())
        for param in chain(self._plan.inputs, self._plan.parameters, self._plan.outputs):
            if param.name in self._values:
                param.actual_value = self._values[param.name]
                values_keys.discard(param.name)

        # NOTE: we need 2-pass the plan parameters as values can be overridden
        # that should be reflected in the params_map
        params_map = TemplateVariableFormatter.to_map(chain(self._plan.inputs, self._plan.parameters))
        for param in chain(self._plan.inputs, self._plan.parameters, self._plan.outputs):
            if isinstance(param.actual_value, str):
                actual_value_set_before = getattr(param, "_v_actual_value_set", False)
                param.actual_value = self._template_engine.apply(param.actual_value, params_map)
                param._v_actual_value_set = actual_value_set_before

        self.missing_parameters = values_keys

        return self._plan


class CompositePlanValueResolver(ValueResolver):
    """Value resolution class for a ``CompositePlan``.

    Applies values and default_values to a nested workflow.

    Order of precedence is as follows (from lowest to highest):
    - Default value on a parameter
    - Default value on a mapping to the parameter
    - Value passed to a mapping to the parameter
    - Value passed to the parameter
    - Value propagated to a parameter from the source of a ParameterLink
    """

    def __init__(self, plan: CompositePlan, values: Optional[Dict[str, Any]] = None):
        super().__init__(plan, values)

    def apply(self) -> AbstractPlan:
        """Applies values and default_values to a ``CompositePlan``.

        Returns:
            A ``CompositePlan`` with values applied.
        """

        if self._values:
            self._apply_parameters_values()

            for name, step in filter(lambda x: isinstance(x[1], dict), self._values.items()):
                child_workflow = next((w for w in self._plan.plans if w.name == name), None)
                if not child_workflow:
                    raise errors.ChildWorkflowNotFoundError(name, self._plan.name)

                rv = ValueResolver.get(child_workflow, step)
                _ = rv.apply()
                self.missing_parameters.update({f"{name}.{mp}" for mp in rv.missing_parameters})

        # apply defaults
        for mapping in self._plan.mappings:
            self._apply_parameter_defaults(mapping)

        apply_parameter_links(self._plan)

        return self._plan

    def _apply_parameter_defaults(self, mapping: ParameterMapping) -> None:
        """Apply default values to a mapping and contained params if they're not set already.

        Args:
            mapping(ParameterMapping): The mapping to apply values to.
        """

        if not mapping.actual_value_set and mapping.default_value:
            mapping.actual_value = mapping.default_value

            for mapped_to in mapping.mapped_parameters:
                if isinstance(mapped_to, ParameterMapping):
                    self._apply_parameter_defaults(mapped_to)
                else:
                    if not mapped_to.actual_value_set:
                        mapped_to.actual_value = mapping.default_value

    def _apply_parameters_values(self) -> None:
        """Apply values to mappings of a CompositePlan."""
        if self._values is None:
            return
        for k, v in filter(lambda x: not isinstance(x[1], dict), self._values.items()):
            mapping = next((m for m in self._plan.mappings if m.name == k), None)

            if not mapping:
                self.missing_parameters.add(k)
                continue

            mapping.actual_value = v


def apply_parameter_links(workflow: CompositePlan) -> None:
    """Apply values from parameter links.

    Args:
        workflow(CompositePlan): The workflow whose links values should be applied on.
    """
    for link in workflow.links:
        for sink in link.sinks:
            sink.actual_value = link.source.actual_value

    for plan in workflow.plans:
        if isinstance(plan, CompositePlan):
            apply_parameter_links(plan)



#
# Copyright 2018-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Update outdated files created by the "run" command.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.update:update
   :prog: renku update
   :nested: full

Recreating outdated files
~~~~~~~~~~~~~~~~~~~~~~~~~

.. image:: ../../_static/asciicasts/update.delay.gif
   :width: 850
   :alt: Update outdated files


The information about dependencies for each file in a Renku project is stored
in various metadata.

When an update command is executed, Renku looks into the most recent execution
of each workflow (Plan and Activity combination) and checks which one is outdated
(i.e. at least one of its inputs is modified). It generates a minimal
dependency graph for each outdated file stored in the repository. It means
that only the necessary steps will be executed.

Assume that the following history for the file ``H`` exists.

.. code-block:: text

          C---D---E
         /         \
    A---B---F---G---H

The first example shows situation when ``D`` is modified and files ``E`` and
``H`` become outdated.

.. code-block:: text

          C--*D*--(E)
         /          \
    A---B---F---G---(H)

    ** - modified
    () - needs update

In this situation, you can do effectively three things:

* Update all files

  .. code-block:: console

     $ renku update --all

* Update only ``E``

  .. code-block:: console

     $ renku update E

* Update ``E`` and ``H``

  .. code-block:: console

     $ renku update H

.. note:: If there were uncommitted changes then the command fails.
   Check :program:`git status` to see details.

In some cases it may be desirable to avoid updating the renku metadata
and to avoid committing this and any other change in the repository when the update
command is run. If this is the case then you can pass the ``--skip-metadata-update``
flag to ``renku update``.

.. cheatsheet::
   :group: Running
   :command: $ renku update [--all] [<path>...]
   :description: Update outdated output files created by renku run. With
                 <path>'s: Only recreate these files. With --all (default):
                 Update all outdated output files.
   :target: rp

Pre-update checks
~~~~~~~~~~~~~~~~~

In the next example, files ``A`` or ``B`` are modified, hence the majority
of dependent files must be recreated.

.. code-block:: text

            (C)--(D)--(E)
           /            \
    *A*--*B*--(F)--(G)--(H)

To avoid excessive recreation of the large portion of files which could have
been affected by a simple change of an input file, consider specifying a single
file (e.g. ``renku update G``). See also :ref:`cli-status`.

.. _cli-update-with-siblings:

Update siblings
~~~~~~~~~~~~~~~

If a workflow step produces multiple output files, these outputs will be always
updated together.

.. code-block:: text

                   (B)
                  /
    *A*--[step 1]--(C)
                  \
                   (D)

An attempt to update a single file would update its siblings as well.

The following commands will produce the same result.

.. code-block:: console

   $ renku update C
   $ renku update B C D

Ignoring deleted paths
~~~~~~~~~~~~~~~~~~~~~~

The update command will regenerate any deleted files/directories. If you don't
want to regenerate deleted paths, pass ``--ignore-deleted`` to the update
command. You can make this the default behavior by setting
``update_ignore_delete`` config value for a project or globally:

  .. code-block:: console

     $ renku config set [--global] update_ignore_delete True

Note that deleted path always will be regenerated if they have siblings or
downstream dependencies that aren't deleted.

"""

import click
from lazy_object_proxy import Proxy

from renku.command.util import WARNING
from renku.core import errors
from renku.ui.cli.utils.callback import ClickCallback
from renku.ui.cli.utils.plugins import available_workflow_providers


@click.command()
@click.option("--all", "-a", "update_all", is_flag=True, default=False, help="Update all outdated files (default).")
@click.option("--dry-run", "-n", is_flag=True, default=False, help="Show a preview of plans that will be executed.")
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1)
@click.option(
    "provider",
    "-p",
    "--provider",
    default="toil",
    show_default=True,
    type=click.Choice(Proxy(available_workflow_providers), case_sensitive=False),
    help="The workflow engine to use.",
)
@click.option(
    "config", "-c", "--config", metavar="<config file>", help="YAML file containing configuration for the provider."
)
@click.option("-i", "--ignore-deleted", is_flag=True, help="Ignore deleted paths.")
@click.option("--skip-metadata-update", is_flag=True, help="Do not update the metadata store for the execution.")
def update(update_all, dry_run, paths, provider, config, ignore_deleted, skip_metadata_update):
    """Update existing files by rerunning their outdated workflow."""
    from renku.command.format.activity import tabulate_activities
    from renku.command.update import update_command

    communicator = ClickCallback()

    if not paths and not update_all and not dry_run:
        message = f"{WARNING}Updating all outputs could trigger expensive computations. Are you sure?"
        communicator.confirm(message, default=True, abort=True)

    try:
        result = (
            update_command(skip_metadata_update=skip_metadata_update)
            .with_communicator(communicator)
            .build()
            .execute(
                update_all=update_all,
                dry_run=dry_run,
                paths=paths,
                provider=provider,
                config=config,
                ignore_deleted=ignore_deleted,
            )
        )
    except errors.NothingToExecuteError:
        exit(1)
    else:
        if dry_run:
            activities, modified_inputs = result.output
            click.echo(tabulate_activities(activities, modified_inputs))



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service plans show controller."""

from renku.command.workflow import export_workflow_command
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.serializers.workflows import WorkflowPlansExportRequest, WorkflowPlansExportResponseRPC
from renku.ui.service.views import result_response


class WorkflowPlansExportCtrl(ServiceCtrl, RenkuOperationMixin):
    """Controller for plan show endpoint."""

    REQUEST_SERIALIZER = WorkflowPlansExportRequest()
    RESPONSE_SERIALIZER = WorkflowPlansExportResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a workflow plan show controller."""
        self.ctx = WorkflowPlansExportCtrl.REQUEST_SERIALIZER.load(request_data)
        super().__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        result = (
            export_workflow_command()
            .build()
            .execute(
                name_or_id=self.ctx["plan_id"],
                format=self.ctx["format"].value,
                output=None,
                values=self.ctx["values"],
                basedir=".",
                resolve_paths=False,
                nest_workflows=True,
            )
        )
        return result.output

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        result = self.execute_op()
        return result_response(WorkflowPlansExportCtrl.RESPONSE_SERIALIZER, result)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git APi provider interface."""

from abc import ABC
from pathlib import Path
from typing import List, Optional, Union


class IGitAPIProvider(ABC):
    """Interface a Git API Provider."""

    def download_files_from_api(
        self,
        files: List[Union[Path, str]],
        folders: List[Union[Path, str]],
        target_folder: Union[Path, str],
        remote: str,
        token: str,
        branch: Optional[str] = None,
    ):
        """Download files through a remote Git API."""
        raise NotImplementedError()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent run templates."""

import marshmallow

from renku.command.schema.agent import PersonSchema
from renku.command.schema.annotation import AnnotationSchema
from renku.command.schema.calamus import JsonLDSchema, Nested, fields, oa, prov, renku, schema
from renku.command.schema.parameter import CommandInputSchema, CommandOutputSchema, CommandParameterSchema
from renku.domain_model.workflow.plan import Plan

MAX_GENERATED_NAME_LENGTH = 25


class PlanSchema(JsonLDSchema):
    """Plan schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Plan, schema.Action, schema.CreativeWork, renku.Plan]
        model = Plan
        unknown = marshmallow.EXCLUDE

    command = fields.String(renku.command, load_default=None)
    description = fields.String(schema.description, load_default=None)
    creators = Nested(schema.creator, PersonSchema, many=True)
    id = fields.Id()
    inputs = Nested(renku.hasInputs, CommandInputSchema, many=True, load_default=None)
    date_created = fields.DateTime(schema.dateCreated, format="iso")
    date_modified = fields.DateTime(schema.dateModified, format="iso")
    date_removed = fields.DateTime(prov.invalidatedAtTime, format="iso")
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)
    name = fields.String(schema.name, load_default=None)
    derived_from = fields.IRI(prov.wasDerivedFrom, load_default=None)
    project_id = fields.IRI(renku.hasPlan, reverse=True)
    outputs = Nested(renku.hasOutputs, CommandOutputSchema, many=True, load_default=None)
    parameters = Nested(renku.hasArguments, CommandParameterSchema, many=True, load_default=None)
    success_codes = fields.List(renku.successCodes, fields.Integer(), load_default=[0])
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)

    @marshmallow.pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            obj.date_created = obj.date_created.replace(microsecond=0)
            obj.date_modified = obj.date_modified.replace(microsecond=0)
            if obj.date_removed:
                obj.date_removed = obj.date_removed.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks needed to determine integrity of datasets."""

import itertools

import click

from renku.command.command_builder import inject
from renku.command.util import WARNING
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.util import communication
from renku.domain_model.project_context import project_context


@inject.autoparams("activity_gateway")
def check_migrated_activity_ids(fix, activity_gateway: IActivityGateway, **_):
    """Check that activity ids were correctly migrated in the past."""
    activities = activity_gateway.get_all_activities(include_deleted=True)

    wrong_activities = [a for a in activities if not a.id.startswith("/activities/")]

    if fix:
        current_database = project_context.database
        for activity in wrong_activities:
            communication.info(f"Fixing activity '{activity.id}'")

            activity_gateway.remove(activity, keep_reference=False)

            # NOTE: Modify id on activity and children
            activity.unfreeze()
            activity.id = f"/activities/{activity.id}"
            activity._p_oid = current_database.hash_id(activity.id)
            activity.freeze()

            for attribute in itertools.chain(
                activity.usages, activity.hidden_usages, activity.generations, activity.parameters
            ):
                object.__setattr__(attribute, "id", f"/activities/{attribute.id}")  # type: ignore

            activity.association.id = f"/activities/{activity.association.id}"

            activity_gateway.add(activity)

        wrong_activities = []

    if not wrong_activities:
        return True, False, None

    problems = (
        WARNING
        + "There are invalid activity ids in the project (use 'renku doctor --fix' to fix them):"
        + "\n\n\t"
        + "\n\t".join(click.style(a.id, fg="yellow") for a in wrong_activities)
        + "\n"
    )

    return False, True, problems


@inject.autoparams("activity_gateway")
def check_activity_dates(fix, activity_gateway: IActivityGateway, **_):
    """Check activities have correct start/end/delete dates.

    Args:
        fix(bool): Whether to fix found issues.
        activity_gateway(IActivityGateway): Injected ActivityGateway.
        _: keyword arguments.

    Returns:
        Tuple[bool, Optional[str]]: Tuple of whether there are activities with invalid dates, if they can be
            automatically fixed and a string of the problem.
    """
    invalid_activities = []

    for activity in activity_gateway.get_all_activities(include_deleted=True):
        plan = activity.association.plan
        if (
            activity.started_at_time < plan.date_created
            or activity.ended_at_time < activity.started_at_time
            or (activity.invalidated_at and activity.invalidated_at < activity.ended_at_time)
        ):
            invalid_activities.append(activity)

    if not invalid_activities:
        return True, False, None
    if not fix:
        ids = [a.id for a in invalid_activities]
        message = (
            WARNING
            + "The following activity have incorrect start, end, or delete date (use 'renku doctor --fix' to fix them):"
            + "\n\t"
            + "\n\t".join(ids)
        )
        return False, True, message

    fix_activity_dates(activities=invalid_activities)
    project_context.database.commit()
    communication.info("Activity dates were fixed")

    return True, False, None


def fix_activity_dates(activities):
    """Fix activities' start/end/delete dates."""
    for activity in activities:
        plan = activity.association.plan
        activity.unfreeze()
        if activity.started_at_time < plan.date_created:
            activity.started_at_time = plan.date_created

        if activity.ended_at_time < activity.started_at_time:
            activity.ended_at_time = activity.started_at_time

        if activity.invalidated_at and activity.invalidated_at < activity.ended_at_time:
            activity.invalidated_at = activity.ended_at_time
        activity.freeze()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template management commands."""

from renku.command.command_builder.command import Command
from renku.core.template.usecase import list_templates, set_template, show_template, update_template, validate_templates


def list_templates_command():
    """Command to list available templates."""
    return Command().command(list_templates)


def show_template_command():
    """Command to show template details."""
    return Command().command(show_template).with_database()


def set_template_command():
    """Command to set template for a project."""
    return (
        Command()
        .command(set_template)
        .lock_project()
        .require_migration()
        .require_clean()
        .with_database(write=True)
        .with_commit()
    )


def update_template_command():
    """Command to update project's template."""
    return (
        Command()
        .command(update_template)
        .lock_project()
        .require_migration()
        .require_clean()
        .with_database(write=True)
        .with_commit()
    )


def validate_templates_command():
    """Command to validate a template repository."""
    return Command().command(validate_templates)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Click utilities."""

from typing import TYPE_CHECKING, List

import click

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderParameter


def shell_complete_datasets(ctx, param, incomplete) -> List[str]:
    """Shell completion for dataset names."""
    from renku.command.dataset import search_datasets_command

    try:
        result = search_datasets_command().build().execute(name=incomplete)
    except Exception:
        return []
    else:
        return result.output


def shell_complete_workflows(ctx, param, incomplete) -> List[str]:
    """Shell completion for plan names."""
    from renku.command.workflow import search_workflows_command

    try:
        result = search_workflows_command().build().execute(name=incomplete)
    except Exception:
        return []
    else:
        return [n for n in result.output if n.startswith(incomplete)]


def shell_complete_sessions(ctx, param, incomplete) -> List[str]:
    """Shell completion for session names."""
    from renku.command.session import search_sessions_command

    try:
        result = search_sessions_command().build().execute(name=incomplete)
    except Exception:
        return []
    else:
        return result.output


def shell_complete_session_providers(ctx, param, incomplete) -> List[str]:
    """Shell completion for session providers names."""
    from renku.command.session import search_session_providers_command

    try:
        result = search_session_providers_command().build().execute(name=incomplete)
    except Exception:
        return []
    else:
        return result.output


class CaseInsensitiveChoice(click.Choice):
    """Case-insensitive click choice.

    Based on https://github.com/pallets/click/issues/569.
    """

    def convert(self, value, param, ctx):
        """Convert value to its choice value."""
        if value is None:
            return None
        return super().convert(value.lower(), param, ctx)


class MutuallyExclusiveOption(click.Option):
    """Custom option class to allow specifying mutually exclusive options in click commands."""

    def __init__(self, *args, **kwargs):
        mutually_exclusive = sorted(kwargs.pop("mutually_exclusive", []))
        self.mutually_exclusive = set()
        self.mutually_exclusive_names = []

        for mutex in mutually_exclusive:
            if type(mutex) == tuple:
                self.mutually_exclusive.add(mutex[0])
                self.mutually_exclusive_names.append(mutex[1])
            else:
                self.mutually_exclusive.add(mutex)
                self.mutually_exclusive_names.append(mutex)

        _help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive_names)
            kwargs["help"] = f"{_help} NOTE: This argument is mutually exclusive with arguments: [{ex_str}]."
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        """Handles the parse result for the option."""
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(self.name, ", ".join(sorted(self.mutually_exclusive_names)))
            )

        return super().handle_parse_result(ctx, opts, args)


def create_options(providers, parameter_function: str):
    """Create options for a group of providers."""

    def wrapper(f):
        from click_option_group import optgroup

        for i, provider in enumerate(sorted(providers, reverse=True, key=lambda p: p.name.lower())):
            parameters: List["ProviderParameter"] = getattr(provider, parameter_function)()
            for j, param in enumerate(parameters):
                param_help = f"\b\n{param.help}\n " if j == 0 else param.help  # NOTE: add newline after a group

                args = (
                    [f"-{a}" if len(a) == 1 else f"--{a}" for a in param.flags if a] + [param.name.replace("-", "_")]
                    if param.flags
                    else [f"--{param.name}"]
                )

                f = optgroup.option(
                    *args,
                    type=param.type,
                    help=param_help,
                    is_flag=param.is_flag,
                    default=param.default,
                    multiple=param.multiple,
                    metavar=param.metavar,
                )(f)

            name = f"{provider.name} configuration"
            if i == len(providers) - 1:
                name = "\n  " + name  # NOTE: add newline before first group

            f = optgroup.group(name=name)(f)

        return f

    return wrapper



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration models V7."""

import os

from marshmallow import EXCLUDE

from renku.command.schema.calamus import fields, prov, renku, schema
from renku.core.migration.utils import OLD_METADATA_PATH, get_datasets_path
from renku.core.util import yaml

from .v3 import Base, DatasetFileSchemaV3, DatasetSchemaV3, UrlSchemaV3


class Dataset(Base):
    """Dataset migration model."""

    @classmethod
    def from_yaml(cls, path, commit=None):
        """Read content from YAML file."""
        data = yaml.read_yaml(path)
        self = DatasetSchemaV7(commit=commit).load(data)
        self._metadata_path = path
        return self

    def to_yaml(self, path=None):
        """Write content to a YAML file."""
        data = DatasetSchemaV7().dump(self)
        path = path or self._metadata_path or os.path.join(self.path, OLD_METADATA_PATH)
        yaml.write_yaml(path=path, data=data)


class DatasetFileSchemaV7(DatasetFileSchemaV3):
    """DatasetFile schema."""

    based_on = fields.Nested(schema.isBasedOn, "DatasetFileSchemaV7", load_default=None)
    source = fields.String(renku.source, load_default=None)


class DatasetSchemaV7(DatasetSchemaV3):
    """Dataset schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Dataset
        model = Dataset
        unknown = EXCLUDE

    derived_from = fields.Nested(prov.wasDerivedFrom, UrlSchemaV3, load_default=None)
    files = fields.Nested(schema.hasPart, DatasetFileSchemaV7, many=True)


def get_project_datasets():
    """Return Dataset migration models for a project."""
    paths = get_datasets_path().rglob(OLD_METADATA_PATH)
    return [Dataset.from_yaml(path=path) for path in paths]



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git APi provider interface."""

import tarfile
import tempfile
from pathlib import Path
from typing import List, Optional, Union

import gitlab

from renku.core import errors
from renku.core.util.os import delete_dataset_file
from renku.domain_model.git import GitURL
from renku.ui.service.interfaces.git_api_provider import IGitAPIProvider


class GitlabAPIProvider(IGitAPIProvider):
    """GitLab API provider abstraction layer.

    Args:
        paths: List of files to download.
        target_folder: Folder to use to download the files.
        remote: Remote repository URL.
        token: User bearer token.
        ref: optional reference to checkout,
    Raises:
        errors.ProjectNotFound: If the remote URL is not accessible.
        errors.AuthenticationError: If the bearer token is invalid in any way.
    """

    def download_files_from_api(
        self,
        files: List[Union[Path, str]],
        folders: List[Union[Path, str]],
        target_folder: Union[Path, str],
        remote: str,
        token: str,
        branch: Optional[str] = None,
    ):
        """Download files through a remote Git API.

        Args:
            files(List[Union[Path, str]]): Files to download.
            folders(List[Union[Path, str]]): Folders to download.
            target_folder(Union[Path, str]): Destination to save downloads to.
            remote(str): Git remote URL.
            token(str): Gitlab API token.
            branch(Optional[str]): Git reference (Default value = None).
        """
        if not branch:
            branch = "HEAD"

        target_folder = Path(target_folder)

        git_data = GitURL.parse(remote)
        try:
            gl = gitlab.Gitlab(git_data.instance_url, oauth_token=token)
            project = gl.projects.get(f"{git_data.owner}/{git_data.name}")
        except gitlab.GitlabAuthenticationError:
            # NOTE: Invalid or expired tokens fail even on public projects. Let's give it a try without tokens
            try:
                gl = gitlab.Gitlab(git_data.instance_url)
                project = gl.projects.get(f"{git_data.owner}/{git_data.name}")
            except gitlab.GitlabAuthenticationError as e:
                raise errors.AuthenticationError from e
            except gitlab.GitlabGetError as e:
                # NOTE: better to re-raise this as a core error since it's a common case
                if "project not found" in getattr(e, "error_message", "").lower():
                    raise errors.ProjectNotFound from e
                else:
                    raise

        for file in files:
            full_path = target_folder / file

            full_path.parent.mkdir(parents=True, exist_ok=True)

            try:
                with open(full_path, "wb") as f:
                    project.files.raw(file_path=str(file), ref=branch, streamed=True, action=f.write)
            except gitlab.GitlabGetError:
                delete_dataset_file(full_path)
                continue

        for folder in folders:
            with tempfile.NamedTemporaryFile() as f:
                project.repository_archive(path=str(folder), sha=branch, streamed=True, action=f.write, format="tar.gz")
                f.seek(0)
                with tarfile.open(fileobj=f) as archive:
                    archive.extractall(path=target_folder)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template view model."""

from typing import Any, Dict, List, Optional

from renku.core.template.template import FileAction, RenderedTemplate, get_sorted_actions
from renku.core.util.util import to_string
from renku.domain_model.template import Template, TemplateParameter


class TemplateViewModel:
    """A view model for a ``Template``."""

    def __init__(
        self,
        aliases: List[str],
        description: str,
        icon: Optional[str],
        id: str,
        immutable_files: Optional[List[str]],
        name: str,
        parameters: List[TemplateParameter],
        reference: Optional[str],
        source: Optional[str],
        version: Optional[str],
        versions: List[str],
    ):
        self.aliases: List[str] = aliases
        self.description: str = description
        self.icon = icon
        self.id: str = id
        self.immutable_files: Optional[List[str]] = immutable_files
        self.name: str = name
        self.parameters: List[TemplateParameterViewModel] = [
            TemplateParameterViewModel.from_template_parameter(p) for p in parameters
        ]
        self.reference = reference
        self.source = source
        self.version = version
        self.versions = versions

    @classmethod
    def from_template(cls, template: Template) -> "TemplateViewModel":
        """Create view model from ``Template``.

        Args:
            template(Template): The input template.

        Returns:
            TemplateViewModel: View model for a template.
        """
        return cls(
            aliases=template.aliases,
            description=template.description,
            icon=template.icon,
            id=template.id,
            immutable_files=template.immutable_files,
            name=template.name,
            parameters=template.parameters,
            reference=template.reference,
            source=template.source,
            version=template.version,
            versions=template.get_all_references(),
        )


class TemplateParameterViewModel:
    """A view model for a ``TemplateParameter``."""

    def __init__(
        self,
        name: str,
        description: str,
        type: str,
        possible_values: List[Any],
        default: Optional[Any],
    ):
        self.name: str = name
        self.description: str = description
        self.type: str = type
        self.possible_values: List[Any] = possible_values
        self.default: Optional[Any] = default

    @classmethod
    def from_template_parameter(cls, parameter: TemplateParameter) -> "TemplateParameterViewModel":
        """Create view model from ``Template``.

        Args:
            parameter(TemplateParameter): The input template parameter.

        Returns:
            TemplateParameterViewModel: View model for a template parameter.
        """
        return cls(
            name=parameter.name,
            description=parameter.description,
            type=to_string(parameter.type),
            possible_values=parameter.possible_values or [],
            default=parameter.default,
        )


class TemplateChangeViewModel:
    """A view model for resulting changes from a template set/update."""

    def __init__(
        self,
        file_changes: List[str],
        id: str,
        old_id: Optional[str],
        reference: Optional[str],
        source: Optional[str],
        version: Optional[str],
    ):
        self.file_changes = file_changes
        self.id: str = id
        self.old_id: Optional[str] = old_id if old_id != id else ""
        self.reference = reference
        self.source = source
        self.version = version

    @classmethod
    def from_template(
        cls, template: RenderedTemplate, actions: Dict[str, FileAction], old_id: Optional[str] = None
    ) -> "TemplateChangeViewModel":
        """Create view model from ``Template``.

        Args:
            template(RenderedTemplate): Input rendered template.
            actions(Dict[str, FileAction]): Mapping of paths to actions taken when rendering the template.
            old_id(Optional[str]: Current template Id.

        Returns:
            TemplateChangeViewModel: View model for the template change.
        """
        actions_mapping: Dict[FileAction, str] = {
            FileAction.APPEND: "Append to",
            FileAction.CREATE: "Initialize",
            FileAction.DELETED: "Ignore deleted file",
            FileAction.IGNORE_IDENTICAL: "Ignore unchanged file",
            FileAction.IGNORE_UNCHANGED_REMOTE: "Ignore unchanged template file",
            FileAction.KEEP: "Keep",
            FileAction.OVERWRITE: "Overwrite",
            FileAction.RECREATE: "Recreate deleted file",
            FileAction.UPDATE_DOCKERFILE: "Update",
        }

        file_changes = [
            f"{actions_mapping[action]} {relative_path} ..."
            for relative_path, action in get_sorted_actions(actions=actions).items()
        ]

        return cls(
            file_changes=file_changes,
            id=template.template.id,
            old_id=old_id,
            reference=template.template.reference,
            source=template.template.source,
            version=template.template.version,
        )



# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service version controller."""
from marshmallow import Schema, fields


class VersionResponse(Schema):
    """Version response schema."""

    latest_version = fields.String()
    supported_project_version = fields.Number()
    minimum_api_version = fields.String()
    maximum_api_version = fields.String()


class VersionResponseRPC(Schema):
    """Version response RPC schema."""

    result = fields.Nested(VersionResponse)



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Communicator class for service communication."""

from renku.core.util.communication import CommunicationCallback


class ServiceCallback(CommunicationCallback):
    """CommunicationCallback implementation for service messages."""

    def __init__(self, user_job=None):
        """Create a new ``ServiceCallback``."""
        super().__init__()

        self.messages = []
        self.warnings = []
        self.errors = []
        self._user_job = user_job

    def echo(self, msg, end="\n"):
        """Write a message."""
        self.messages.append(msg)

    def info(self, msg):
        """Write an info message."""
        self.messages.append(msg)

    def warn(self, msg):
        """Write a warning message."""
        self.warnings.append(msg)

    def error(self, msg):
        """Write an error message."""
        self.errors.append(msg)

    def confirm(self, msg, abort=False, warning=False, default=False):
        """Get confirmation for an action using a prompt."""
        return False

    def start_progress(self, name, total, **kwargs):
        """Start job tracking."""
        self._user_job.extras = {
            "description": name,
            "total_size": total,
        }

    def update_progress(self, name, amount):
        """Update job status."""
        self._user_job.extras["progress_size"] = amount
        self._user_job.save()

    def finalize_progress(self, name):
        """End job tracking."""
        self._user_job.complete()
